from datetime import timedelta
from typing import Any

import pytest
from pydantic import BaseModel, ConfigDict
from temporalio.client import WorkflowFailureError

from mistralai.workflows import activity, get_workflow_definition, workflow
from mistralai.workflows.exceptions import WorkflowsException

from .fixtures import (
    GreetingParams,
    ParamsModel,
    UpdateInput,
    WorkflowWithPydanticHandlers,
)
from .utils import create_test_worker


class SimpleInput(BaseModel):
    message: str


class ComplexOutput(BaseModel):
    result: str
    count: int


class PromptInput(BaseModel):
    model_config = ConfigDict(extra="forbid")
    prompt: str


class CountInput(BaseModel):
    model_config = ConfigDict(extra="forbid")
    count: int


@workflow.define(name="workflow_with_union_input")
class WorkflowWithUnionInput:
    @workflow.entrypoint
    async def run(self, params: PromptInput | CountInput) -> str:
        if isinstance(params, PromptInput):
            return f"prompt:{params.prompt}"
        return f"count:{params.count}"


@workflow.define(name="workflow_with_optional_union_input")
class WorkflowWithOptionalUnionInput:
    @workflow.entrypoint
    async def run(self, params: PromptInput | CountInput | None) -> str:
        if params is None:
            return "none"
        if isinstance(params, PromptInput):
            return f"prompt:{params.prompt}"
        return f"count:{params.count}"


@workflow.define(name="workflow_with_defaults")
class WorkflowWithDefaults:
    @workflow.entrypoint
    async def run(self, name: str, greeting: str = "Hello", count: int = 1) -> str:
        return f"{greeting}, {name}! (x{count})"


@workflow.define(name="workflow_all_defaults")
class WorkflowAllDefaults:
    @workflow.entrypoint
    async def run(self, name: str = "World", greeting: str = "Hello") -> str:
        return f"{greeting}, {name}!"


@workflow.define(name="workflow_mixed_params")
class WorkflowMixedParams:
    @workflow.entrypoint
    async def run(
        self,
        required_str: str,
        optional_int: int = 42,
        optional_bool: bool = False,
        optional_str: str = "default",
    ) -> str:
        return f"{required_str}: int={optional_int}, bool={optional_bool}, str={optional_str}"


@activity()
async def activity_with_defaults(name: str, prefix: str = "Mr.", suffix: str = "") -> str:
    result = f"{prefix} {name}"
    if suffix:
        result += f" {suffix}"
    return result


@activity()
async def activity_all_defaults(x: int = 10, y: int = 20) -> int:
    return x + y


@workflow.define(name="workflow_calling_activity_with_defaults")
class WorkflowCallingActivityWithDefaults:
    @workflow.entrypoint
    async def run(self, name: str) -> str:
        result1 = await activity_with_defaults(name, "Dr.", "PhD")
        result2 = await activity_with_defaults(name, "Ms.")
        result3 = await activity_with_defaults(name)
        result4 = await activity_all_defaults()
        result5 = await activity_all_defaults(5)
        result6 = await activity_all_defaults(5, 15)
        return f"{result1} | {result2} | {result3} | {result4} | {result5} | {result6}"


@workflow.define(name="workflow_with_signal_defaults")
class WorkflowWithSignalDefaults:
    def __init__(self) -> None:
        self.messages: list[str] = []

    @workflow.entrypoint
    async def run(self) -> str:
        await workflow.wait_condition(lambda: len(self.messages) >= 2)
        return " | ".join(self.messages)

    @workflow.signal(name="add_message")
    async def add_message(self, text: str, priority: int = 0, tag: str = "info") -> None:
        self.messages.append(f"[{tag}:{priority}] {text}")


@workflow.define(name="workflow_with_query_defaults")
class WorkflowWithQueryDefaults:
    def __init__(self) -> None:
        self.counter = 0

    @workflow.entrypoint
    async def run(self) -> str:
        await workflow.wait_condition(lambda: self.counter >= 10)
        return "done"

    @workflow.signal(name="increment")
    async def increment(self, amount: int = 1) -> None:
        self.counter += amount

    @workflow.query(name="get_counter")
    def get_counter(self, multiplier: int = 1, offset: int = 0) -> int:
        return (self.counter * multiplier) + offset


@workflow.define(name="workflow_with_update_defaults")
class WorkflowWithUpdateDefaults:
    def __init__(self) -> None:
        self.value = 100

    @workflow.entrypoint
    async def run(self) -> int:
        await workflow.wait_condition(lambda: self.value <= 0)
        return self.value

    @workflow.update(name="adjust_value")
    async def adjust_value(self, delta: int, min_value: int = 0, max_value: int = 200) -> dict:
        old_value = self.value
        self.value += delta
        self.value = max(min_value, min(max_value, self.value))
        return {"old": old_value, "new": self.value, "clamped": old_value + delta != self.value}


@workflow.define(name="test_signal_empty_dict_defaults")
class SignalEmptyDictDefaultsWorkflow:
    def __init__(self) -> None:
        self.counter = 0

    @workflow.entrypoint
    async def run(self) -> int:
        await workflow.wait_condition(lambda: self.counter >= 2)
        return self.counter

    @workflow.signal(name="increment")
    async def increment(self, amount: int = 1) -> None:
        self.counter += amount


class TestWorkflowDefaultParameters:
    @pytest.mark.asyncio
    async def test_workflow_with_some_defaults(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowWithDefaults], activities=[]):
            workflow_def = get_workflow_definition(WorkflowWithDefaults)
            assert workflow_def is not None

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"name": "Alice", "greeting": "Hi", "count": 3},
                id="test-all-params",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "Hi, Alice! (x3)"

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"name": "Bob"},
                id="test-required-only",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "Hello, Bob! (x1)"

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"name": "Charlie", "count": 5},
                id="test-partial-defaults",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "Hello, Charlie! (x5)"

    @pytest.mark.asyncio
    async def test_workflow_with_all_defaults(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowAllDefaults], activities=[]):
            workflow_def = get_workflow_definition(WorkflowAllDefaults)
            assert workflow_def is not None

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {},
                id="test-no-params",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "Hello, World!"

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"name": "Python"},
                id="test-one-param",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "Hello, Python!"

    @pytest.mark.asyncio
    async def test_workflow_mixed_default_types(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowMixedParams], activities=[]):
            workflow_def = get_workflow_definition(WorkflowMixedParams)
            assert workflow_def is not None

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"required_str": "test"},
                id="test-mixed-required-only",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "test: int=42, bool=False, str=default"

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"required_str": "custom", "optional_int": 99, "optional_bool": True},
                id="test-mixed-partial",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "custom: int=99, bool=True, str=default"


class TestActivityDefaultParameters:
    @pytest.mark.asyncio
    async def test_activity_with_defaults(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[WorkflowCallingActivityWithDefaults],
            activities=[activity_with_defaults, activity_all_defaults],
        ):
            workflow_def = get_workflow_definition(WorkflowCallingActivityWithDefaults)
            assert workflow_def is not None

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"name": "Smith"},
                id="test-activity-defaults",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            expected = "Dr. Smith PhD | Ms. Smith | Mr. Smith | 30 | 25 | 20"
            assert result["result"] == expected


class TestSignalDefaultParameters:
    @pytest.mark.asyncio
    async def test_signal_with_defaults(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowWithSignalDefaults], activities=[]):
            workflow_def = get_workflow_definition(WorkflowWithSignalDefaults)
            assert workflow_def is not None

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {},
                id="test-signal-defaults",
                task_queue="test-task-queue",
            )

            await handle.signal("add_message", {"text": "urgent", "priority": 10, "tag": "error"})

            await handle.signal("add_message", {"text": "normal"})

            result = await handle.result()
            assert "[error:10] urgent" in result["result"]
            assert "[info:0] normal" in result["result"]

    @pytest.mark.asyncio
    async def test_signal_empty_dict_uses_all_defaults(self, temporal_env: Any) -> None:
        """
        Regression test: empty dict with all default parameters.

        Without selective @functools.wraps (excluding __annotations__), Temporal would see
        the original function signature and fail to deserialize an empty dict.
        See workflows/worker/workflow.py signal/query/update decorators.
        """
        async with create_test_worker(temporal_env, workflows=[SignalEmptyDictDefaultsWorkflow], activities=[]):
            handle = await temporal_env.client.start_workflow(
                "test_signal_empty_dict_defaults",
                {},
                id="test-signal-empty-dict",
                task_queue="test-task-queue",
            )

            await handle.signal("increment", {"amount": 1})
            await handle.signal("increment", {})

            result = await handle.result()
            assert result["result"] == 2


class TestQueryDefaultParameters:
    @pytest.mark.asyncio
    async def test_query_with_defaults(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowWithQueryDefaults], activities=[]):
            workflow_def = get_workflow_definition(WorkflowWithQueryDefaults)
            assert workflow_def is not None

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {},
                id="test-query-defaults",
                task_queue="test-task-queue",
            )

            await handle.signal("increment", {"amount": 5})
            await handle.signal("increment", {})  # default amount=1

            result = await handle.query("get_counter", {})
            assert result == 6  # 5 + 1, multiplier=1, offset=0

            result = await handle.query("get_counter", {"multiplier": 2})
            assert result == 12  # (5 + 1) * 2 + 0

            result = await handle.query("get_counter", {"multiplier": 3, "offset": 10})
            assert result == 28  # (5 + 1) * 3 + 10


class TestUpdateDefaultParameters:
    @pytest.mark.asyncio
    async def test_update_with_defaults(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowWithUpdateDefaults], activities=[]):
            workflow_def = get_workflow_definition(WorkflowWithUpdateDefaults)
            assert workflow_def is not None

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {},
                id="test-update-defaults",
                task_queue="test-task-queue",
            )

            result = await handle.execute_update("adjust_value", {"delta": -30})
            assert result["old"] == 100
            assert result["new"] == 70
            assert result["clamped"] is False

            result = await handle.execute_update("adjust_value", {"delta": 150, "max_value": 150})
            assert result["old"] == 70
            assert result["new"] == 150
            assert result["clamped"] is True

            result = await handle.execute_update("adjust_value", {"delta": -200, "min_value": -10})
            assert result["old"] == 150
            assert result["new"] == -10
            assert result["clamped"] is True


class TestPydanticParameters:
    @pytest.mark.asyncio
    async def test_pydantic_signal_query_update_handlers(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowWithPydanticHandlers], activities=[]):
            workflow_def = get_workflow_definition(WorkflowWithPydanticHandlers)
            assert workflow_def is not None

            params = ParamsModel(name="TestWorkflow", count=2)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                params.model_dump(),
                id="test-pydantic-handlers",
                task_queue="test-task-queue",
            )

            query_result = await handle.query("get_status")
            assert isinstance(query_result, dict)
            assert "message" in query_result
            assert "TestWorkflow" in query_result["message"]

            update_input = UpdateInput(new_value="updated_value")
            update_result = await handle.execute_update("update_value", update_input.model_dump())
            assert isinstance(update_result, dict)
            assert update_result["old_value"] == "TestWorkflow"
            assert update_result["new_value"] == "updated_value"
            assert update_result["success"] is True

            await handle.signal("add_greeting", GreetingParams(name="Alice").model_dump())
            await handle.signal("add_greeting", GreetingParams(name="Bob").model_dump())

            result = await handle.result()
            assert isinstance(result, dict)
            assert result["success"] is True
            assert "2 greetings" in result["message"]


class TestUnionInputWorkflow:
    @pytest.mark.asyncio
    async def test_prompt_input(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowWithUnionInput], activities=[]):
            workflow_def = get_workflow_definition(WorkflowWithUnionInput)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"prompt": "hello"},
                id="test-union-prompt",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "prompt:hello"

    @pytest.mark.asyncio
    async def test_count_input(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowWithUnionInput], activities=[]):
            workflow_def = get_workflow_definition(WorkflowWithUnionInput)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"count": 42},
                id="test-union-count",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "count:42"

    @pytest.mark.asyncio
    async def test_wrong_input(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowWithUnionInput], activities=[]):
            workflow_def = get_workflow_definition(WorkflowWithUnionInput)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"count": 42, "wrong": "key"},
                id="test-union-wrong-input",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=1),
            )
            with pytest.raises(WorkflowFailureError):
                await handle.result()


class TestOptionalUnionInputWorkflow:
    @pytest.mark.asyncio
    async def test_prompt_input(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowWithOptionalUnionInput], activities=[]):
            workflow_def = get_workflow_definition(WorkflowWithOptionalUnionInput)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"prompt": "hello"},
                id="test-optional-union-prompt",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "prompt:hello"

    @pytest.mark.asyncio
    async def test_count_input(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowWithOptionalUnionInput], activities=[]):
            workflow_def = get_workflow_definition(WorkflowWithOptionalUnionInput)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"count": 7},
                id="test-optional-union-count",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "count:7"

    @pytest.mark.asyncio
    async def test_wrong_input(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowWithOptionalUnionInput], activities=[]):
            workflow_def = get_workflow_definition(WorkflowWithOptionalUnionInput)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"count": 7, "wrong": "key"},
                id="test-optional-union-wrong-input",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=1),
            )
            with pytest.raises(WorkflowFailureError):
                await handle.result()

    @pytest.mark.asyncio
    async def test_none_input(self, temporal_env: Any) -> None:
        """Passing None (null payload) should resolve to params=None in the handler."""
        async with create_test_worker(temporal_env, workflows=[WorkflowWithOptionalUnionInput], activities=[]):
            workflow_def = get_workflow_definition(WorkflowWithOptionalUnionInput)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                None,
                id="test-optional-union-none",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == "none"


# -- Workflows for input variety tests --


@workflow.define(name="workflow_no_params")
class WorkflowNoParams:
    @workflow.entrypoint
    async def run(self) -> str:
        return "no_params"


@workflow.define(name="workflow_single_int_default")
class WorkflowSingleIntDefault:
    @workflow.entrypoint
    async def run(self, value: int = 0) -> str:
        return f"value:{value}"


@workflow.define(name="workflow_single_str_default")
class WorkflowSingleStrDefault:
    @workflow.entrypoint
    async def run(self, text: str = "") -> str:
        return f"text:{text}"


@workflow.define(name="workflow_single_bool_default")
class WorkflowSingleBoolDefault:
    @workflow.entrypoint
    async def run(self, flag: bool = False) -> str:
        return f"flag:{flag}"


@workflow.define(name="workflow_single_required_str")
class WorkflowSingleRequiredStr:
    @workflow.entrypoint
    async def run(self, name: str) -> str:
        return f"name:{name}"


@workflow.define(name="workflow_single_basemodel")
class WorkflowSingleBaseModel:
    @workflow.entrypoint
    async def run(self, params: SimpleInput) -> str:
        return f"msg:{params.message}"


class TestNoParamWorkflow:
    @pytest.mark.asyncio
    async def test_no_params_no_arg(self, temporal_env: Any) -> None:
        """Workflow with no params, started with no arg."""
        async with create_test_worker(temporal_env, workflows=[WorkflowNoParams], activities=[]):
            workflow_def = get_workflow_definition(WorkflowNoParams)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                id="test-no-params-no-arg",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "no_params"

    @pytest.mark.asyncio
    async def test_no_params_empty_dict(self, temporal_env: Any) -> None:
        """Workflow with no params, started with {}."""
        async with create_test_worker(temporal_env, workflows=[WorkflowNoParams], activities=[]):
            workflow_def = get_workflow_definition(WorkflowNoParams)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {},
                id="test-no-params-empty-dict",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "no_params"

    @pytest.mark.asyncio
    async def test_no_params_none(self, temporal_env: Any) -> None:
        """Workflow with no params, started with None."""
        async with create_test_worker(temporal_env, workflows=[WorkflowNoParams], activities=[]):
            workflow_def = get_workflow_definition(WorkflowNoParams)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                None,
                id="test-no-params-none",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == "no_params"


class TestFalsyDefaultInputs:
    """Test workflows where the default parameter value is falsy (0, '', False).

    These are started with the falsy value explicitly to verify it is not
    accidentally coerced to {}.
    """

    @pytest.mark.asyncio
    async def test_int_zero_via_dict(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleIntDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleIntDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"value": 0},
                id="test-int-zero-dict",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "value:0"

    @pytest.mark.asyncio
    async def test_int_nonzero_via_dict(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleIntDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleIntDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"value": 42},
                id="test-int-nonzero-dict",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "value:42"

    @pytest.mark.asyncio
    async def test_int_default_via_empty_dict(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleIntDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleIntDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {},
                id="test-int-default-empty-dict",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "value:0"

    @pytest.mark.asyncio
    async def test_empty_string_via_dict(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleStrDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleStrDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"text": ""},
                id="test-empty-str-dict",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "text:"

    @pytest.mark.asyncio
    async def test_nonempty_string_via_dict(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleStrDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleStrDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"text": "hello"},
                id="test-nonempty-str-dict",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "text:hello"

    @pytest.mark.asyncio
    async def test_false_bool_via_dict(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleBoolDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleBoolDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"flag": False},
                id="test-false-bool-dict",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "flag:False"

    @pytest.mark.asyncio
    async def test_true_bool_via_dict(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleBoolDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleBoolDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"flag": True},
                id="test-true-bool-dict",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "flag:True"


class TestRawPositionalArg:
    """Test workflows started with a raw positional arg (not a dict).

    Temporal's start_workflow(name, arg, ...) passes `arg` directly to the
    workflow wrapper. The conversion layer must handle non-dict values.
    """

    @pytest.mark.asyncio
    async def test_int_zero_as_raw_arg(self, temporal_env: Any) -> None:
        """Falsy int 0 passed as raw arg must not break the workflow."""
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleIntDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleIntDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                0,
                id="test-raw-int-zero",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == "value:0"

    @pytest.mark.asyncio
    async def test_int_nonzero_as_raw_arg(self, temporal_env: Any) -> None:
        """Truthy int passed as raw arg."""
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleIntDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleIntDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                42,
                id="test-raw-int-nonzero",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == "value:42"

    @pytest.mark.asyncio
    async def test_empty_string_as_raw_arg(self, temporal_env: Any) -> None:
        """Falsy empty string passed as raw arg."""
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleStrDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleStrDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                "",
                id="test-raw-empty-str",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == "text:"

    @pytest.mark.asyncio
    async def test_nonempty_string_as_raw_arg(self, temporal_env: Any) -> None:
        """Truthy string passed as raw arg."""
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleStrDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleStrDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                "hello",
                id="test-raw-nonempty-str",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == "text:hello"

    @pytest.mark.asyncio
    async def test_false_as_raw_arg(self, temporal_env: Any) -> None:
        """Falsy False passed as raw arg."""
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleBoolDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleBoolDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                False,
                id="test-raw-false",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == "flag:False"

    @pytest.mark.asyncio
    async def test_true_as_raw_arg(self, temporal_env: Any) -> None:
        """Truthy True passed as raw arg."""
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleBoolDefault], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleBoolDefault)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                True,
                id="test-raw-true",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == "flag:True"

    @pytest.mark.asyncio
    async def test_string_as_raw_arg_for_required_param(self, temporal_env: Any) -> None:
        """Required string param passed as raw arg."""
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleRequiredStr], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleRequiredStr)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                "Alice",
                id="test-raw-required-str",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == "name:Alice"


class TestSingleRequiredParam:
    @pytest.mark.asyncio
    async def test_string_param_via_dict(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleRequiredStr], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleRequiredStr)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"name": "Alice"},
                id="test-required-str-dict",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "name:Alice"


class TestBaseModelParam:
    @pytest.mark.asyncio
    async def test_basemodel_via_dict(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleBaseModel], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleBaseModel)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"message": "hello"},
                id="test-basemodel-dict",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "msg:hello"

    @pytest.mark.asyncio
    async def test_basemodel_via_model_dump(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[WorkflowSingleBaseModel], activities=[]):
            workflow_def = get_workflow_definition(WorkflowSingleBaseModel)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                SimpleInput(message="world").model_dump(),
                id="test-basemodel-model-dump",
                task_queue="test-task-queue",
            )
            result = await handle.result()
            assert result["result"] == "msg:world"


class PageState(BaseModel):
    offset: int = 0
    total: int = 0


class PageResult(BaseModel):
    total: int
    done: bool = True


@activity()
async def fetch_page(offset: int) -> list[str]:
    if offset >= 200:
        return []
    return [f"item_{i}" for i in range(offset, min(offset + 100, 200))]


@workflow.define(name="workflow_continue_as_new_recursive")
class WorkflowContinueAsNewRecursive:
    @workflow.entrypoint
    async def run(self, state: PageState) -> PageResult:
        items = await fetch_page(state.offset)
        if not items:
            return PageResult(total=state.total)
        next_state = PageState(offset=state.offset + 100, total=state.total + len(items))
        workflow.continue_as_new(next_state)


class TestContinueAsNewBaseModelInstance:
    @pytest.mark.asyncio
    async def test_continue_as_new_with_basemodel(self, temporal_env: Any) -> None:
        """workflow.continue_as_new(PageState(...)) restarts the workflow with fresh history.
        The conversion layer must handle BaseModel params correctly."""
        async with create_test_worker(
            temporal_env,
            workflows=[WorkflowContinueAsNewRecursive],
            activities=[fetch_page],
        ):
            workflow_def = get_workflow_definition(WorkflowContinueAsNewRecursive)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                PageState(offset=100, total=100).model_dump(),
                id="test-continue-as-new-basemodel",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["total"] == 200
            assert result["done"] is True


# -- continue-as-new workflows for various parameter shapes --


class IntCounterState(BaseModel):
    count: int = 0


@workflow.define(name="can_int_counter")
class ContinueAsNewIntCounter:
    """Single int param with default, carried via BaseModel state."""

    @workflow.entrypoint
    async def run(self, state: IntCounterState) -> int:
        if state.count >= 3:
            return state.count
        workflow.continue_as_new(IntCounterState(count=state.count + 1))


class StrAccumulatorState(BaseModel):
    text: str


@workflow.define(name="can_str_accumulator")
class ContinueAsNewStrAccumulator:
    """Single required str param."""

    @workflow.entrypoint
    async def run(self, state: StrAccumulatorState) -> str:
        if len(state.text) >= 5:
            return state.text
        workflow.continue_as_new(StrAccumulatorState(text=state.text + "x"))


class BoolFlipperState(BaseModel):
    flag: bool = False


@workflow.define(name="can_bool_flipper")
class ContinueAsNewBoolFlipper:
    """Single bool param with default. Continues-as-new once with negated value."""

    @workflow.entrypoint
    async def run(self, state: BoolFlipperState) -> bool:
        if state.flag:
            return state.flag
        workflow.continue_as_new(BoolFlipperState(flag=not state.flag))


class MultiParamState(BaseModel):
    x: int = 0
    y: int = 0
    label: str = "start"


@workflow.define(name="can_multi_param")
class ContinueAsNewMultiParam:
    """Multiple params with defaults."""

    @workflow.entrypoint
    async def run(self, state: MultiParamState) -> str:
        if state.x >= 2:
            return f"{state.label}:{state.x}:{state.y}"
        workflow.continue_as_new(MultiParamState(x=state.x + 1, y=state.y + 10, label=f"{state.label}+"))


@workflow.define(name="can_union_input")
class ContinueAsNewUnionInput:
    """Union param. Continues-as-new with a different variant."""

    @workflow.entrypoint
    async def run(self, params: PromptInput | CountInput) -> str:
        if isinstance(params, CountInput):
            return f"count:{params.count}"
        workflow.continue_as_new(CountInput(count=len(params.prompt)))


class TestContinueAsNewParameterVariety:
    """Test that workflow.continue_as_new works with various parameter shapes.

    Each workflow uses the proper continue-as-new API (not recursive self.run)
    to restart with fresh history and new parameters.
    """

    @pytest.mark.asyncio
    async def test_int_with_initial_value(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[ContinueAsNewIntCounter], activities=[]):
            workflow_def = get_workflow_definition(ContinueAsNewIntCounter)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                IntCounterState(count=1).model_dump(),
                id="test-can-int",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == 3

    @pytest.mark.asyncio
    async def test_int_with_default(self, temporal_env: Any) -> None:
        """Start with default count=0."""
        async with create_test_worker(temporal_env, workflows=[ContinueAsNewIntCounter], activities=[]):
            workflow_def = get_workflow_definition(ContinueAsNewIntCounter)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                IntCounterState().model_dump(),
                id="test-can-int-default",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == 3

    @pytest.mark.asyncio
    async def test_str_required(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[ContinueAsNewStrAccumulator], activities=[]):
            workflow_def = get_workflow_definition(ContinueAsNewStrAccumulator)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                StrAccumulatorState(text="ab").model_dump(),
                id="test-can-str",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == "abxxx"

    @pytest.mark.asyncio
    async def test_bool_with_default(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[ContinueAsNewBoolFlipper], activities=[]):
            workflow_def = get_workflow_definition(ContinueAsNewBoolFlipper)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                BoolFlipperState(flag=False).model_dump(),
                id="test-can-bool",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] is True

    @pytest.mark.asyncio
    async def test_multi_param(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[ContinueAsNewMultiParam], activities=[]):
            workflow_def = get_workflow_definition(ContinueAsNewMultiParam)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                MultiParamState(x=0, y=0, label="go").model_dump(),
                id="test-can-multi",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == "go++:2:20"

    @pytest.mark.asyncio
    async def test_basemodel(self, temporal_env: Any) -> None:
        """BaseModel with activity calls between continue-as-new iterations."""
        async with create_test_worker(
            temporal_env,
            workflows=[WorkflowContinueAsNewRecursive],
            activities=[fetch_page],
        ):
            workflow_def = get_workflow_definition(WorkflowContinueAsNewRecursive)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                PageState(offset=100, total=100).model_dump(),
                id="test-can-basemodel",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["total"] == 200

    @pytest.mark.asyncio
    async def test_union(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[ContinueAsNewUnionInput], activities=[]):
            workflow_def = get_workflow_definition(ContinueAsNewUnionInput)
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                PromptInput(prompt="hello").model_dump(),
                id="test-can-union",
                task_queue="test-task-queue",
                execution_timeout=timedelta(seconds=5),
            )
            result = await handle.result()
            assert result["result"] == "count:5"


class TestInvalidUnionDefinition:
    def test_mixed_union_raises_at_definition_time(self) -> None:
        with pytest.raises(WorkflowsException, match="unsupported union members: str"):

            @workflow.define(name="workflow_with_invalid_union")
            class _WorkflowWithInvalidUnion:
                @workflow.entrypoint
                async def run(self, params: PromptInput | str) -> str:
                    return ""



import os
from typing import Generator

import httpx
import pytest


@pytest.fixture(autouse=True)
def setup_test_config() -> Generator[None, None, None]:
    # Override the autouse fixture from workflow_sdk/tests/conftest.py, which sets
    # config.worker.deployment_name = "test-task-queue" for unit tests. Integration
    # tests run against a real server, so the config must reflect the actual env vars
    # (DEPLOYMENT_NAME=example-dev-worker) rather than the unit-test sentinel value.
    yield


@pytest.fixture
def api_client() -> httpx.AsyncClient:
    server_url = os.getenv("SERVER_URL", "http://localhost:7444")
    api_key = os.getenv("MISTRAL_API_KEY")

    return httpx.AsyncClient(
        base_url=server_url,
        headers={
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
        },
        timeout=120.0,
    )



import inspect
from datetime import timedelta
from types import NoneType
from typing import TYPE_CHECKING, Any, Literal, Type, overload

import structlog
import temporalio.workflow
from pydantic import BaseModel
from temporalio.workflow import ChildWorkflowHandle as TemporalChildWorkflowHandle
from temporalio.workflow import ParentClosePolicy

from mistralai.workflows.core.definition.validation.validator import get_function_signature_type_hints
from mistralai.workflows.core.definition.workflow_definition import (
    _get_workflow_entrypoint_method,
    get_workflow_definition,
)
from mistralai.workflows.core.dependencies.dependency_injector import DependencyInjector
from mistralai.workflows.core.utils.contextvars import unwrap_contextual_result
from mistralai.workflows.exceptions import ErrorCode, WorkflowsException

if TYPE_CHECKING:
    from typing import NoReturn

    from typing_extensions import deprecated

    class ChildWorkflowHandle(TemporalChildWorkflowHandle):
        """Typing-only subclass that hides Temporal-internal attributes."""

        @property
        @deprecated("temporal internal")
        def first_execution_run_id(self) -> NoReturn: ...


logger = structlog.get_logger(__name__)


def _coerce_workflow_result(return_value: Any, entrypoint_method: Any) -> Any:
    """Coerce a raw workflow result based on the entrypoint's return-type annotation.

    Handles two cases:
    - BaseModel return types are validated via model_validate
    - Primitive return types are extracted from the {"result": ...} Temporal wrapper
    """
    original_func = getattr(entrypoint_method, "__wrapped__", None)
    if original_func:
        _, return_type, _ = get_function_signature_type_hints(original_func, is_method=True)
    else:
        return_type = None

    if return_type and return_type != NoneType:
        try:
            is_base_model = inspect.isclass(return_type) and issubclass(return_type, BaseModel)
        except TypeError:
            is_base_model = False

        if is_base_model:
            return return_type.model_validate(return_value)
        elif isinstance(return_value, dict) and "result" in return_value:
            return return_value["result"]

    return return_value


_HIDDEN_ATTRS = frozenset({"first_execution_run_id"})


def _patch_child_workflow_handle(
    handle: TemporalChildWorkflowHandle,
    entrypoint_method: Any,
) -> None:
    """Patch a Temporal ChildWorkflowHandle in-place:

    1. Monkey-patch ``_resolve_success`` so that ``await handle`` returns the
       same unwrapped + coerced value as the ``wait=True`` path.
    2. Monkey-patch ``__getattribute__`` to hide Temporal-internal attributes
       (e.g. ``first_execution_run_id``) from consumers.

    NOTE: Relies on Temporal's private ``_resolve_success`` method.
    """
    original_resolve = handle._resolve_success  # type: ignore[attr-defined]

    def _coercing_resolve(result: Any) -> None:
        _, unwrapped = unwrap_contextual_result(result)
        original_resolve(_coerce_workflow_result(unwrapped, entrypoint_method))

    handle._resolve_success = _coercing_resolve  # type: ignore[attr-defined]

    original_getattribute = type(handle).__getattribute__

    def _guarded_getattribute(self: Any, name: str) -> Any:
        if name in _HIDDEN_ATTRS:
            raise AttributeError(f"'{type(self).__name__}' has no attribute '{name}'")
        return original_getattribute(self, name)

    handle.__class__ = type(
        handle.__class__.__name__,
        (handle.__class__,),
        {"__getattribute__": _guarded_getattribute},
    )


def get_execution_id() -> str | None:
    if temporalio.workflow.in_workflow():
        return temporalio.workflow.info().workflow_id
    else:
        return None


@overload
async def execute_workflow(
    workflow: Type,
    params: BaseModel,
    execution_timeout: timedelta = ...,
    execution_id: str | None = ...,
    wait: Literal[True] = ...,
    parent_close_policy: ParentClosePolicy | None = ...,
) -> Any: ...


@overload
async def execute_workflow(
    workflow: Type,
    params: BaseModel,
    execution_timeout: timedelta = ...,
    execution_id: str | None = ...,
    wait: Literal[False] = ...,
    parent_close_policy: ParentClosePolicy | None = ...,
) -> "ChildWorkflowHandle": ...


@overload
async def execute_workflow(
    workflow: Type,
    params: BaseModel,
    execution_timeout: timedelta = ...,
    execution_id: str | None = ...,
    wait: bool = ...,
    parent_close_policy: ParentClosePolicy | None = ...,
) -> Any: ...


async def execute_workflow(
    workflow: Type,
    params: BaseModel,
    execution_timeout: timedelta = timedelta(hours=1),
    execution_id: str | None = None,
    wait: bool = True,
    parent_close_policy: ParentClosePolicy | None = None,
) -> Any:
    """Execute a workflow. If called from within a workflow, it will execute the workflow as a child workflow.

    Args:
        workflow (Type): The workflow class to execute. (must be decorated with @define)
        params (BaseModel): The parameters to pass to the workflow. (must be a BaseModel)
        execution_timeout (timedelta, optional): The maximum time the workflow can run.
                                                 Defaults to timedelta(hours=1).
        execution_id (str | None, optional): The workflow id to use. If None, a random id will be generated.
        wait (bool, optional): If True (default), wait for the child workflow to complete and return the result.
                               If False, return a ChildWorkflowHandle immediately (only inside a workflow).
        parent_close_policy (ParentClosePolicy | None, optional): Policy for the child when the parent closes.
                             Defaults to TERMINATE when wait=True, ABANDON when wait=False.

    Returns:
        Any: The return value of the workflow (wait=True), or a ChildWorkflowHandle (wait=False).
    """
    if not wait and not temporalio.workflow.in_workflow():
        raise WorkflowsException(
            code=ErrorCode.WORKFLOW_DEFINITION_ERROR,
            message="execute_workflow with wait=False can only be called from within a workflow",
        )

    workflow_definition = get_workflow_definition(workflow)
    if not workflow_definition:
        raise WorkflowsException(
            code=ErrorCode.WORKFLOW_DEFINITION_ERROR,
            message=f"{workflow} class must be decorated with @workflow.define",
        )

    entrypoint_method = _get_workflow_entrypoint_method(workflow)
    if not entrypoint_method:
        raise WorkflowsException(
            code=ErrorCode.WORKFLOW_DEFINITION_ERROR,
            message=f"{workflow} class must have an entrypoint method decorated with @workflow.entrypoint",
        )

    if parent_close_policy is None:
        parent_close_policy = ParentClosePolicy.TERMINATE if wait else ParentClosePolicy.ABANDON

    if temporalio.workflow.in_workflow():
        if not wait:
            handle = await temporalio.workflow.start_child_workflow(
                workflow_definition.name,
                params.model_dump(),
                execution_timeout=execution_timeout,
                result_type=dict,
                id=execution_id,
                parent_close_policy=parent_close_policy,
            )
            _patch_child_workflow_handle(handle, entrypoint_method)
            return handle

        return_value = await temporalio.workflow.execute_child_workflow(
            workflow_definition.name,
            params.model_dump(),
            execution_timeout=execution_timeout,
            result_type=dict,
            id=execution_id,
            parent_close_policy=parent_close_policy,
        )

        _, return_value = unwrap_contextual_result(return_value)
    else:
        workflow_instance = workflow()
        params_dict = params.model_dump()

        if DependencyInjector.is_inside_dependencies_context():
            return_value = await entrypoint_method(workflow_instance, params_dict)
        else:
            dependency_injector = DependencyInjector.get_singleton_instance()
            async with dependency_injector.with_dependencies():
                return_value = await entrypoint_method(workflow_instance, params_dict)

    return _coerce_workflow_result(return_value, entrypoint_method)



import os
import socket
from enum import StrEnum
from typing import Self

import structlog
from pydantic import AliasChoices, Field, SecretStr, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from mistralai.workflows.core.logging import Env, LogFormat, LoggerConfig, LogLevel, setup_logging
from mistralai.workflows.core.rate_limiting.rate_limit import RateLimit
from mistralai.workflows.core.storage.blob_storage_impl import StorageProvider
from mistralai.workflows.exceptions import ErrorCode, WorkflowsException
from mistralai.workflows.protocol.v1.workflow import DeploymentName

logger = structlog.getLogger(__name__)

env_file = ".env" if os.environ.get("PYTEST_VERSION") is None else ".env.test"


class CommonConfig(BaseSettings):
    app_name: str = "mistral-workflows"
    app_version: str = "0.0.0"
    env: Env = Env.DEV
    log_format: LogFormat = LogFormat.CONSOLE
    log_level: LogLevel = LogLevel.INFO
    enable_docs: bool = Field(default=True, validation_alias="ENABLE_DOCS")

    otel_enabled: bool = False
    otel_endpoint: str = "http://localhost:4318"
    otel_sample_rate: float = 1.0
    otel_export_interval_ms: int = 30000
    otel_tail_sampling: bool = False
    otel_local: bool = False
    otel_inject_logs: bool = True

    ca_bundle: str | None = Field(  # type: ignore[pydantic-alias]
        default=None,
        description="Path to the CA bundle file for TLS verification",
        validation_alias=AliasChoices("CA_BUNDLE", "CURL_CA_BUNDLE"),
    )

    mistral_api_key: SecretStr | None = Field(default=None)

    model_config = SettingsConfigDict(
        env_file=env_file,
        env_file_encoding="utf-8",
        extra="ignore",
        env_parse_none_str="null",
    )


# Hardcoded reserved workflow names that conflict with API endpoints
# This CANNOT be overridden via environment variables for security reasons
RESERVED_WORKFLOW_NAMES: frozenset[str] = frozenset({"executions", "schedules", "definitions", "internal"})

RESERVED_UPDATE_NAMES: frozenset[str] = frozenset({"__submit_input"})

RESERVED_QUERY_NAMES: frozenset[str] = frozenset({"__get_pending_inputs"})

INTERNAL_ACTIVITY_PREFIX: str = "__internal__"

RESERVED_INPUT_ATTRIBUTE_PREFIX: str = "__internal_"


class ReservedExtraFieldsAttribute(StrEnum):
    """Reserved attribute names injected via extra_fields (pydantic extras).

    These attributes are injected by the system and should not be used by the user.
    All values must be prefixed with RESERVED_INPUT_ATTRIBUTE_PREFIX.
    """

    EXPERIMENTAL_IDENTITY = f"{RESERVED_INPUT_ATTRIBUTE_PREFIX}experimental_identity"


class TemporalConfig(BaseSettings):
    server_url: str = Field(
        default="localhost:7233",
    )
    namespace: str = Field(
        default="mistral-workflows",
    )
    api_key: SecretStr | None = Field(
        default=None,
    )
    namespace_retention_days: int = Field(
        default=30,
    )
    single_tenant_namespace: bool = Field(
        default=False,
        description="If True, the namespace matches the customer/workspace identifier.",
    )
    enforce_payload_encoding: bool = Field(
        default=False,
    )
    task_queue: str = Field(default="default")  # Allow override
    tls: bool = False
    external_server_url: str | None = None
    http_proxy_target_host: str | None = Field(
        default=None,
        description="Target host for the HTTP CONNECT proxy (e.g. 'proxy.example.com:8080')",
    )
    http_proxy_basic_auth_user: str | None = Field(
        default=None,
        description="Basic auth username for the HTTP CONNECT proxy",
    )
    http_proxy_basic_auth_pass: SecretStr | None = Field(
        default=None,
        description="Basic auth password for the HTTP CONNECT proxy",
    )

    model_config = SettingsConfigDict(
        env_file=env_file,
        env_file_encoding="utf-8",
        extra="ignore",
        env_prefix="TEMPORAL_",
        env_parse_none_str="null",
    )

    @model_validator(mode="after")
    def set_external_url(self) -> "Self":
        if not self.external_server_url:
            self.external_server_url = self.server_url
        return self


class BlobStorageConfig(BaseSettings):
    enabled: bool = False
    storage_provider: StorageProvider = StorageProvider.S3
    prefix: str | None = None

    # Azure settings
    container_name: str | None = None
    azure_connection_string: SecretStr | None = None

    # GCS settings
    bucket_id: str | None = None

    # S3 settings
    bucket_name: str | None = None
    region_name: str | None = None
    endpoint_url: str | None = None
    aws_access_key_id: SecretStr | None = None
    aws_secret_access_key: SecretStr | None = None

    model_config = SettingsConfigDict(
        env_file=env_file,
        env_file_encoding="utf-8",
        extra="ignore",
        env_parse_none_str="null",
    )


class PayloadOffloadingConfig(BaseSettings):
    enabled: bool = False
    storage_config: BlobStorageConfig | None = None
    min_size_bytes: int = 1024 * 1024  # 1MB


class PayloadEncryptionMode(StrEnum):
    NONE = "none"
    FULL = "full"
    PARTIAL = "partial"


class PayloadEncryptionConfig(BaseSettings):
    mode: PayloadEncryptionMode = PayloadEncryptionMode.NONE

    # If both keys are provided, the main key will be used for encryption and both keys will be used for decryption
    # to support key rotation.
    main_key: SecretStr | None = None
    secondary_key: SecretStr | None = None


class AgentConfig(BaseSettings):
    llm_rate_limit: RateLimit | None = None
    mistral_client_server: str | None = None
    mistral_client_server_url: str | None = None
    mistral_client_url_params: dict[str, str] | None = None
    mistral_client_timeout_ms: int | None = None
    mistral_client_api_key: SecretStr | None = Field(  # type: ignore[pydantic-alias]
        default=None,
        description="API key for Mistral client",
        validation_alias=AliasChoices("MISTRAL_CLIENT_API_KEY", "MISTRAL_API_KEY"),
    )
    model_config = SettingsConfigDict(
        env_file=env_file,
        env_file_encoding="utf-8",
        extra="ignore",
        env_parse_none_str="null",
        env_nested_delimiter="__",
    )


class WorkerVersioningConfig(BaseSettings):
    enabled: bool = Field(default=False)
    # Temporal deployment name used for WorkerDeploymentVersion. This is the name that
    # Temporal uses to match pollers to registered versions. In controller mode it comes
    # from TEMPORAL_DEPLOYMENT_NAME ({namespace}/{resource-name}); in manual mode from
    # DEPLOYMENT_NAME. This is independent of WorkerConfig.deployment_name which is used
    # for the heartbeat/internal Workflows API.
    deployment_name: str | None = Field(default=None)
    build_id: str | None = Field(
        default=None,
        validation_alias="BUILD_ID",
    )
    auto_register_as_current: bool | None = Field(
        default=None,
        validation_alias="WORKER_AUTO_REGISTER_AS_CURRENT",
    )

    @model_validator(mode="after")
    def configure_versioning(self) -> "WorkerVersioningConfig":
        """Infer worker versioning behaviour from environment and config values."""
        controller_deployment = os.environ.get("TEMPORAL_DEPLOYMENT_NAME")
        controller_build_id = os.environ.get("TEMPORAL_WORKER_BUILD_ID")
        is_managed_by_controller = bool(controller_deployment and controller_build_id)

        auto_register_override = self.auto_register_as_current

        if is_managed_by_controller:
            # Use the controller's deployment name (TEMPORAL_DEPLOYMENT_NAME) for Temporal
            # versioning. The controller registers versions under this name, so the worker
            # must poll under the same name for Temporal to match pollers to versions.
            self.deployment_name = controller_deployment
            self.build_id = controller_build_id
            self.enabled = True
            if self.auto_register_as_current is None:
                self.auto_register_as_current = False
            logger.info(
                "Worker managed by Temporal Worker Controller",
                deployment_name=self.deployment_name,
                build_id=self.build_id,
                auto_register=self.auto_register_as_current,
            )
            return self

        # manual / local mode
        if self.deployment_name and self.build_id:
            self.enabled = True
            if auto_register_override is not None:
                self.auto_register_as_current = auto_register_override
            else:
                self.auto_register_as_current = True

            logger.info(
                "Worker in manual mode with versioning enabled",
                deployment_name=self.deployment_name,
                build_id=self.build_id,
                auto_register=self.auto_register_as_current,
            )
            return self

        if self.enabled:
            logger.warning(
                "Worker versioning enabled but deployment name or build ID missing; disabling versioning",
                deployment_name=self.deployment_name,
                build_id=self.build_id,
            )

        self.enabled = False
        self.auto_register_as_current = False
        return self

    model_config = SettingsConfigDict(
        env_file=env_file,
        env_file_encoding="utf-8",
        extra="ignore",
        env_parse_none_str="null",
    )


class WorkerConfig(BaseSettings):
    retry_policy_max_attempts: int = 3
    retry_policy_backoff_coefficient: float = 2.0
    # - The workflow fails on any workflow-level error if set to True. Note: Workflow-level errors differ from
    # activity-level errors and typically result from bugs in the workflow code. By default, workflows do not fail
    #  on errors, enabling you to push code fixes without re-running workflows.
    # More information: https://community.temporal.io/t/workflow-retry-policy-seems-to-not-be-getting-respected/11203/2",
    dangerously_force_fail_workflow_on_error: bool = Field(
        default=False,
        description="☢️ DANGER ZONE ☢️",
        alias="DANGEROUSLY_FORCE_FAIL_WORKFLOW_ON_ERROR",
    )

    server_url: str = Field(  # type: ignore[pydantic-alias]
        default="https://api.mistral.ai",
        validation_alias=AliasChoices("SERVER_URL", "server_url"),
    )
    api_version: str = "v1"
    allow_multiple_workers: bool = True
    enable_config_discovery: bool = True
    mistral_api_headers: dict[str, str] | None = None

    temporal_payload_offloading: PayloadOffloadingConfig = Field(default_factory=PayloadOffloadingConfig)
    temporal_payload_encryption: PayloadEncryptionConfig = Field(default_factory=PayloadEncryptionConfig)

    activity_attributes_offloading: PayloadOffloadingConfig = Field(default_factory=PayloadOffloadingConfig)

    agent: AgentConfig = Field(default_factory=AgentConfig)
    versioning: WorkerVersioningConfig = Field(default_factory=WorkerVersioningConfig)

    workflow_name_prefix: str = Field(default="", description="If set, all workflows will be prefixed with this value.")
    default_enforce_determinism: bool = Field(
        default=True,
        description="If True, all workflows will be sandboxed by default to enforce deterministic execution.",
    )

    deployment_name: DeploymentName = None
    worker_name: str = Field(default_factory=socket.gethostname)

    health_server_host: str = "localhost"
    health_server_port: int | None = None

    model_config = SettingsConfigDict(
        env_file=env_file,
        env_file_encoding="utf-8",
        extra="ignore",
        env_parse_none_str="null",
        env_nested_delimiter="__",
    )

    @model_validator(mode="after")
    def warn_determinism_disabled(self) -> "WorkerConfig":
        if not self.default_enforce_determinism:
            logger.warning(
                "Worker configured with default determinism enforcement disabled. "
                "This may lead to non-deterministic behavior and replay failures.",
            )
        return self


class SentryConfig(BaseSettings):
    dsn: SecretStr | None = None

    model_config = SettingsConfigDict(
        env_file=env_file,
        env_file_encoding="utf-8",
        extra="ignore",
        env_prefix="SENTRY_",
        env_parse_none_str="null",
    )


class AppConfig(BaseSettings):
    common: CommonConfig = Field(default_factory=CommonConfig)
    worker: WorkerConfig = Field(default_factory=WorkerConfig)
    sentry: SentryConfig = Field(default_factory=SentryConfig)
    blob_storage: BlobStorageConfig = Field(default_factory=BlobStorageConfig)
    temporal: TemporalConfig = Field(default_factory=TemporalConfig)

    model_config = SettingsConfigDict(
        env_file=env_file,
        env_file_encoding="utf-8",
        extra="ignore",
        env_parse_none_str="null",
    )

    @property
    def otel_headers(self) -> str | None:
        """Generate OTEL headers with Mistral API key as bearer token if available."""
        if self.common.mistral_api_key and self.common.mistral_api_key.get_secret_value():
            return f"Authorization=Bearer {self.common.mistral_api_key.get_secret_value()}"
        return None

    def get_effective_task_queue(self, raise_on_conflict: bool = False) -> str:
        deployment_name = self.worker.deployment_name
        task_queue = self.temporal.task_queue

        if not deployment_name:
            return task_queue
        if task_queue == "default" or task_queue == deployment_name:
            return deployment_name
        if not raise_on_conflict:
            return deployment_name
        raise WorkflowsException(
            code=ErrorCode.WORKER_RUNTIME_CONFIG_ERROR,
            message=(
                f"TEMPORAL_TASK_QUEUE ({task_queue!r}) conflicts with DEPLOYMENT_NAME ({deployment_name!r}). "
                "They must be the same value, or leave TEMPORAL_TASK_QUEUE unset "
                "to let it be derived from DEPLOYMENT_NAME."
            ),
        )

    @model_validator(mode="after")
    def inject_defaults(self) -> Self:
        has_temporal_key = self.temporal.api_key and self.temporal.api_key.get_secret_value()
        has_mistral_key = self.common.mistral_api_key and self.common.mistral_api_key.get_secret_value()
        if not has_temporal_key and has_mistral_key:
            self.temporal.api_key = self.common.mistral_api_key
        return self


def _get_or_load_config() -> AppConfig:
    """Read and initialize the application configuration."""
    # Load configuration from environment/file
    config = AppConfig()

    # Set OTEL_EXPORTER_OTLP_CERTIFICATE if CA_BUNDLE is set to enable TLS verification for OpenTelemetry
    if config.common.ca_bundle is not None:
        os.environ["OTEL_EXPORTER_OTLP_CERTIFICATE"] = config.common.ca_bundle

    # Set OTEL_EXPORTER_OTLP_HEADERS if Mistral API key is available
    if config.otel_headers is not None:
        os.environ["OTEL_EXPORTER_OTLP_HEADERS"] = config.otel_headers

    # Set up structured logging with OpenTelemetry trace injection if enabled
    if not structlog.is_configured():
        setup_logging(
            log_level=config.common.log_level,
            log_format=config.common.log_format,
            app_version=config.common.app_version,
            inject_otel_trace=config.common.otel_enabled and config.common.otel_inject_logs,
            extra_config=[
                # noisy
                LoggerConfig(
                    name="httpx",
                    level=LogLevel.WARNING,
                ),
                LoggerConfig(
                    name="httpcore",
                    level=LogLevel.WARNING,
                ),
                LoggerConfig(
                    name="asyncio",
                    level=LogLevel.WARNING,
                ),
                LoggerConfig(
                    name="aiocache",
                    level=LogLevel.WARNING,
                ),
                LoggerConfig(
                    name="urllib3",
                    level=LogLevel.WARNING,
                ),
            ],
        )

    logger.info("Configuration loaded", config=config.model_dump())
    return config


config = _get_or_load_config()



from typing import Any

import pytest
from temporalio.client import WorkflowFailureError
from temporalio.exceptions import ApplicationError, ChildWorkflowError

import mistralai.workflows as workflows
from mistralai.workflows import get_workflow_definition
from mistralai.workflows.exceptions import WorkflowsException

from .fixtures import (
    ChildWorkflow,
    ChildWorkflowCustomExecute,
    ChildWorkflowParams,
    CompareWaitModesParentWorkflow,
    CompareWaitModesPydanticParentWorkflow,
    FailingChildWaitFalseParent,
    FailingChildWaitTrueParent,
    FailingWorkflow,
    FireAndForgetChildParams,
    FireAndForgetChildWorkflow,
    FireAndForgetParentWorkflow,
    MultiArgChildWorkflow,
    MultiArgParentWorkflow,
    MultiParamWithPrefixInput,
    NestedChildLevel1,
    NestedChildLevel2,
    NestedParentWorkflow,
    ParentWorkflow,
    PersonData,
    PrimitiveChildWorkflow,
    PrimitiveParentWorkflow,
    PrimitiveWorkflowParams,
    ProcessingResult,
    PydanticChildWorkflow,
    PydanticParentWorkflow,
    StandaloneWorkflowWithMultiArgs,
    StandaloneWorkflowWithPydantic,
    StartThenAwaitParentWorkflow,
)
from .utils import create_test_worker


class TestExecuteWorkflowBasic:
    @pytest.mark.asyncio
    async def test_simple_child_workflow(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[ParentWorkflow, ChildWorkflow], activities=[]):
            workflow_def = get_workflow_definition(ParentWorkflow)
            assert workflow_def is not None
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"name": "Test"},
                id="test-parent-workflow",
                task_queue="test-task-queue",
            )

            result = await handle.result()
            assert "Child says: Test" in result["result"]
            assert "Parent got:" in result["result"]


class TestExecuteWorkflowWithPydantic:
    @pytest.mark.asyncio
    async def test_child_workflow_with_pydantic_model(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env, workflows=[PydanticParentWorkflow, PydanticChildWorkflow], activities=[]
        ):
            workflow_def = get_workflow_definition(PydanticParentWorkflow)
            person = PersonData(first_name="John", last_name="Doe", age=25)

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                person.model_dump(),
                id="test-pydantic-parent",
                task_queue="test-task-queue",
            )

            result = await handle.result()
            assert "Processed: John Doe is an adult" in result["result"]

    @pytest.mark.asyncio
    async def test_child_workflow_with_pydantic_minor(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env, workflows=[PydanticParentWorkflow, PydanticChildWorkflow], activities=[]
        ):
            workflow_def = get_workflow_definition(PydanticParentWorkflow)
            person = PersonData(first_name="Jane", last_name="Smith", age=16)

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                person.model_dump(),
                id="test-pydantic-parent-minor",
                task_queue="test-task-queue",
            )

            result = await handle.result()
            assert "Processed: Jane Smith is a minor" in result["result"]


class TestExecuteWorkflowWithMultipleArgs:
    @pytest.mark.asyncio
    async def test_child_workflow_with_multiple_args(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env, workflows=[MultiArgParentWorkflow, MultiArgChildWorkflow], activities=[]
        ):
            workflow_def = get_workflow_definition(MultiArgParentWorkflow)

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"name": "Hello", "count": 3},
                id="test-multi-arg-parent",
                task_queue="test-task-queue",
            )

            result = await handle.result()
            assert "Parent received: Hello repeated 3 times: Hello, Hello, Hello" in result["result"]


class TestExecuteWorkflowWithPrimitives:
    @pytest.mark.asyncio
    async def test_child_workflow_with_primitive_types(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env, workflows=[PrimitiveParentWorkflow, PrimitiveChildWorkflow], activities=[]
        ):
            workflow_def = get_workflow_definition(PrimitiveParentWorkflow)

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"message": "TestMessage"},
                id="test-primitive-parent",
                task_queue="test-task-queue",
            )

            result = await handle.result()
            assert "Got back: Echo: TestMessage" in result["result"]


class TestExecuteWorkflowNested:
    @pytest.mark.asyncio
    async def test_nested_child_workflows(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[NestedParentWorkflow, NestedChildLevel1, NestedChildLevel2],
            activities=[],
        ):
            workflow_def = get_workflow_definition(NestedParentWorkflow)

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"value": "test"},
                id="test-nested-parent",
                task_queue="test-task-queue",
            )

            result = await handle.result()
            assert result["result"] == "L0[L1[L2[test]]]"


class TestExecuteWorkflowDirect:
    @pytest.mark.asyncio
    async def test_direct_execute_workflow_with_pydantic(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[StandaloneWorkflowWithPydantic], activities=[]):
            person = PersonData(first_name="Alice", last_name="Johnson", age=30)

            result = await workflows.execute_workflow(StandaloneWorkflowWithPydantic, params=person)

            assert isinstance(result, ProcessingResult)
            assert result.full_name == "Alice Johnson"
            assert result.is_adult is True
            assert "Standalone: Alice Johnson is an adult" in result.message

    @pytest.mark.asyncio
    async def test_direct_execute_workflow_with_multiple_args(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[StandaloneWorkflowWithMultiArgs], activities=[]):
            result = await workflows.execute_workflow(
                StandaloneWorkflowWithMultiArgs, params=MultiParamWithPrefixInput(name="Item", count=3, prefix=">> ")
            )

            assert isinstance(result, str)
            assert result == "Generated: >> Item, >> Item, >> Item"

    @pytest.mark.asyncio
    async def test_direct_execute_workflow_returns_primitive(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[PrimitiveChildWorkflow], activities=[]):
            result = await workflows.execute_workflow(
                PrimitiveChildWorkflow, params=PrimitiveWorkflowParams(message="DirectCall")
            )

            assert isinstance(result, str)
            assert result == "Echo: DirectCall"


class TestExecuteWorkflowNoWait:
    @pytest.mark.asyncio
    async def test_fire_and_forget(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[FireAndForgetParentWorkflow, FireAndForgetChildWorkflow],
            activities=[],
        ):
            workflow_def = get_workflow_definition(FireAndForgetParentWorkflow)
            assert workflow_def is not None
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"tag": "abc"},
                id="test-fire-and-forget-parent",
                task_queue="test-task-queue",
            )

            result = await handle.result()
            assert result["result"] == "parent-done:abc"

    @pytest.mark.asyncio
    async def test_start_then_await(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[StartThenAwaitParentWorkflow, FireAndForgetChildWorkflow],
            activities=[],
        ):
            workflow_def = get_workflow_definition(StartThenAwaitParentWorkflow)
            assert workflow_def is not None
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"tag": "xyz"},
                id="test-start-then-await-parent",
                task_queue="test-task-queue",
            )

            result = await handle.result()
            assert result["result"] == "parent-got:child-done:xyz"

    @pytest.mark.asyncio
    async def test_wait_false_outside_workflow_raises(self) -> None:
        with pytest.raises(WorkflowsException):
            await workflows.execute_workflow(
                FireAndForgetChildWorkflow,
                params=FireAndForgetChildParams(tag="test"),
                wait=False,
            )


class TestExecuteWorkflowWaitComparison:
    """Compare wait=True and wait=False return values from within a workflow."""

    @pytest.mark.asyncio
    async def test_wait_true_and_wait_false_return_same_result(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[CompareWaitModesParentWorkflow, FireAndForgetChildWorkflow],
            activities=[],
        ):
            workflow_def = get_workflow_definition(CompareWaitModesParentWorkflow)
            assert workflow_def is not None
            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"tag": "compare"},
                id="test-compare-wait-modes",
                task_queue="test-task-queue",
            )

            result = await handle.result()
            wait_true_result = result["result"]["wait_true"]
            wait_false_result = result["result"]["wait_false"]

            assert wait_true_result == wait_false_result, (
                f"wait=True and wait=False returned different results!\n"
                f"  wait=True:  {wait_true_result}\n"
                f"  wait=False: {wait_false_result}\n"
            )


class TestFailingChildWorkflow:
    """Verify that failing child workflows surface the same error regardless of wait mode."""

    @staticmethod
    def _get_cause_chain(exc: BaseException) -> list[type]:
        """Return the list of exception types in the __cause__ chain."""
        chain: list[type] = [type(exc)]
        while exc.__cause__ is not None:
            exc = exc.__cause__
            chain.append(type(exc))
        return chain

    @pytest.mark.asyncio
    async def test_failing_child_same_error_both_modes(self, temporal_env: Any) -> None:
        """Ensure the full error chain and message are identical for wait=True and wait=False."""
        errors: dict[str, WorkflowFailureError] = {}

        for label, parent_cls, wf_id in [
            ("wait_true", FailingChildWaitTrueParent, "test-fail-cmp-wait-true"),
            ("wait_false", FailingChildWaitFalseParent, "test-fail-cmp-wait-false"),
        ]:
            async with create_test_worker(
                temporal_env,
                workflows=[parent_cls, FailingWorkflow],
                activities=[],
            ):
                workflow_def = get_workflow_definition(parent_cls)
                h = await temporal_env.client.start_workflow(
                    workflow_def.name,
                    {"message": "same-error"},
                    id=wf_id,
                    task_queue="test-task-queue",
                )
                with pytest.raises(WorkflowFailureError) as exc_info:
                    await h.result()
                errors[label] = exc_info.value

        chain_true = self._get_cause_chain(errors["wait_true"])
        chain_false = self._get_cause_chain(errors["wait_false"])

        assert chain_true == chain_false, (
            f"Different exception chains:\n  wait=True:  {chain_true}\n  wait=False: {chain_false}"
        )
        assert chain_true == [WorkflowFailureError, ChildWorkflowError, ApplicationError]

        root_true = errors["wait_true"]
        root_false = errors["wait_false"]
        while root_true.__cause__ is not None:
            root_true = root_true.__cause__
        while root_false.__cause__ is not None:
            root_false = root_false.__cause__

        assert str(root_true) == str(root_false), (
            f"Different root error messages:\n  wait=True:  {root_true}\n  wait=False: {root_false}"
        )


class TestChildWorkflowWithPydanticParams:
    """Test child workflows that receive a pydantic model return the same result with wait=True and wait=False."""

    @pytest.mark.asyncio
    async def test_pydantic_child_same_result_both_modes(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[CompareWaitModesPydanticParentWorkflow, PydanticChildWorkflow],
            activities=[],
        ):
            workflow_def = get_workflow_definition(CompareWaitModesPydanticParentWorkflow)
            person = PersonData(first_name="Alice", last_name="Johnson", age=30)

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                person.model_dump(),
                id="test-pydantic-child-wait-modes",
                task_queue="test-task-queue",
            )

            result = await handle.result()
            wait_true_result = result["result"]["wait_true"]
            wait_false_result = result["result"]["wait_false"]

            assert wait_true_result == wait_false_result, (
                f"wait=True and wait=False returned different results!\n"
                f"  wait=True:  {wait_true_result}\n"
                f"  wait=False: {wait_false_result}\n"
            )
            assert wait_true_result["full_name"] == "Alice Johnson"
            assert wait_true_result["is_adult"] is True


class TestChildWorkflowCustomEntrypoint:
    def test_child_workflow_with_custom_entrypoint_name_has_correct_method(self) -> None:
        assert hasattr(ChildWorkflowCustomExecute, "execute")
        assert callable(ChildWorkflowCustomExecute.execute)

    @pytest.mark.asyncio
    async def test_execute_workflow_works_with_custom_entrypoint_name(self, temporal_env: Any) -> None:
        async with create_test_worker(temporal_env, workflows=[ChildWorkflowCustomExecute], activities=[]):
            result = await workflows.execute_workflow(
                ChildWorkflowCustomExecute, params=ChildWorkflowParams(name="Test")
            )

            assert result == "Child with custom entrypoint says: Test"



import asyncio
import json
import logging
import uuid
from typing import Any, cast

import httpx
from pydantic import BaseModel, TypeAdapter
from temporalio.client import WorkflowHandle

from mistralai.workflows.protocol.v1.events import (
    CustomTaskCompleted,
    CustomTaskInProgress,
    CustomTaskStarted,
    WorkflowEvent,
    WorkflowEventType,
)
from mistralai.workflows.protocol.v1.streaming import StreamEventSsePayload
from mistralai.workflows.testing.constants import DEFAULT_TEST_TASK_QUEUE

SSE_DATA_PREFIX = "data:"
SSE_RETRY_PREFIX = "retry:"

logger = logging.getLogger(__name__)
sse_payload_adapter: TypeAdapter[StreamEventSsePayload] = TypeAdapter(StreamEventSsePayload)
workflow_event_adapter: TypeAdapter[WorkflowEvent] = TypeAdapter(WorkflowEvent)


async def create_workflow(
    client: httpx.AsyncClient,
    workflow_identifier: str,
    input_data: dict[str, Any] | None = None,
    task_queue: str | None = DEFAULT_TEST_TASK_QUEUE,
) -> dict[str, Any]:
    body: dict[str, Any] = {}
    if input_data is not None:
        body["input"] = input_data
    if task_queue is not None:
        body["task_queue"] = task_queue
    response = await client.post(
        f"/v1/workflows/{workflow_identifier}/execute",
        json=body,
    )
    response.raise_for_status()
    return cast(dict[str, Any], response.json())


async def get_workflow_status(client: httpx.AsyncClient, execution_id: str) -> dict[str, Any]:
    response = await client.get(f"/v1/workflows/executions/{execution_id}")
    response.raise_for_status()
    return cast(dict[str, Any], response.json())


async def get_workflow_definition(client: httpx.AsyncClient, workflow_identifier: str) -> dict[str, Any]:
    response = await client.get(f"/v1/workflows/{workflow_identifier}")
    response.raise_for_status()
    return cast(dict[str, Any], response.json())


async def terminate_workflow(client: httpx.AsyncClient, execution_id: str) -> None:
    response = await client.post(f"/v1/workflows/executions/{execution_id}/terminate")
    response.raise_for_status()


async def cancel_workflow(client: httpx.AsyncClient, execution_id: str) -> httpx.Response:
    return await client.post(f"/v1/workflows/executions/{execution_id}/cancel")


async def signal_workflow(
    client: httpx.AsyncClient, execution_id: str, signal_name: str, signal_input: Any = None
) -> None:
    response = await client.post(
        f"/v1/workflows/executions/{execution_id}/signals",
        json={"name": signal_name, "input": signal_input},
    )
    response.raise_for_status()


async def query_workflow(
    client: httpx.AsyncClient, execution_id: str, query_name: str, query_input: Any = None
) -> dict[str, Any]:
    payload = {"name": query_name}
    if query_input is not None:
        payload["input"] = query_input
    response = await client.post(
        f"/v1/workflows/executions/{execution_id}/queries",
        json=payload,
    )
    response.raise_for_status()
    return cast(dict[str, Any], response.json())


async def update_workflow(
    client: httpx.AsyncClient, execution_id: str, update_name: str, update_input: Any = None
) -> dict[str, Any]:
    payload = {"name": update_name}
    if update_input is not None:
        payload["input"] = update_input
    response = await client.post(
        f"/v1/workflows/executions/{execution_id}/updates",
        json=payload,
    )
    response.raise_for_status()
    return cast(dict[str, Any], response.json())


async def update_workflow_definition(
    client: httpx.AsyncClient, workflow_identifier: str, payload: dict[str, Any]
) -> dict[str, Any]:
    response = await client.put(f"/v1/workflows/{workflow_identifier}", json=payload)
    response.raise_for_status()
    return cast(dict[str, Any], response.json())


async def register_workflow(
    client: httpx.AsyncClient,
    workflow_name: str,
    task_queue: str = "default",
    input_schema: dict[str, Any] | None = None,
) -> dict[str, Any]:
    if input_schema is None:
        input_schema = {"type": "object", "properties": {}, "additionalProperties": True}
    response = await client.post(
        "/v1/workflows/register",
        json={"definitions": [{"name": workflow_name, "task_queue": task_queue, "input_schema": input_schema}]},
    )
    response.raise_for_status()
    return cast(dict[str, Any], response.json())


async def archive_workflow(client: httpx.AsyncClient, workflow_identifier: str) -> httpx.Response:
    response = await client.put(f"/v1/workflows/{workflow_identifier}/archive")
    return response


async def unarchive_workflow(client: httpx.AsyncClient, workflow_identifier: str) -> httpx.Response:
    response = await client.put(f"/v1/workflows/{workflow_identifier}/unarchive")
    return response


async def create_workflow_raw(
    client: httpx.AsyncClient,
    workflow_identifier: str,
    input_data: dict[str, Any],
    task_queue: str | None = DEFAULT_TEST_TASK_QUEUE,
) -> httpx.Response:
    """Start a workflow without raising on error status."""
    body: dict[str, Any] = {"input": input_data}
    if task_queue is not None:
        body["task_queue"] = task_queue
    return await client.post(
        f"/v1/workflows/{workflow_identifier}/execute",
        json=body,
    )


async def signal_workflow_raw(
    client: httpx.AsyncClient, execution_id: str, signal_name: str, signal_input: dict[str, Any] | None = None
) -> httpx.Response:
    """Send a signal without raising on error status."""
    payload: dict[str, Any] = {"name": signal_name}
    if signal_input is not None:
        payload["input"] = signal_input
    return await client.post(
        f"/v1/workflows/executions/{execution_id}/signals",
        json=payload,
    )


async def query_workflow_raw(
    client: httpx.AsyncClient, execution_id: str, query_name: str, query_input: dict[str, Any] | None = None
) -> httpx.Response:
    """Send a query without raising on error status."""
    payload: dict[str, Any] = {"name": query_name}
    if query_input is not None:
        payload["input"] = query_input
    return await client.post(
        f"/v1/workflows/executions/{execution_id}/queries",
        json=payload,
    )


async def update_workflow_raw(
    client: httpx.AsyncClient, execution_id: str, update_name: str, update_input: dict[str, Any] | None = None
) -> httpx.Response:
    """Send an update without raising on error status."""
    payload: dict[str, Any] = {"name": update_name}
    if update_input is not None:
        payload["input"] = update_input
    return await client.post(
        f"/v1/workflows/executions/{execution_id}/updates",
        json=payload,
    )


async def get_workflow(client: httpx.AsyncClient, workflow_identifier: str) -> dict[str, Any]:
    response = await client.get(f"/v1/workflows/{workflow_identifier}")
    response.raise_for_status()
    return cast(dict[str, Any], response.json())


async def wait_for_workflow_completion(
    client: httpx.AsyncClient, execution_id: str, max_retries: int = 50, delay: float = 0.1
) -> dict[str, Any]:
    for _ in range(max_retries):
        status_response = await get_workflow_status(client, execution_id)
        if status_response["status"] in ["COMPLETED", "FAILED", "TERMINATED", "CANCELED"]:
            return status_response
        await asyncio.sleep(delay)
    raise AssertionError(f"Workflow {execution_id} did not complete after {max_retries} retries")


def build_workflow_event_payload(
    execution_id: str,
    workflow_run_id: str,
    workflow_name: str,
    event_type: str,
    attributes: dict[str, Any],
) -> dict[str, Any]:
    return {
        "event": {
            "event_id": str(uuid.uuid4()),
            "event_timestamp": 1234567890000000000,
            "root_workflow_exec_id": execution_id,
            "workflow_exec_id": execution_id,
            "workflow_run_id": workflow_run_id,
            "workflow_name": workflow_name,
            "event_type": event_type,
            "attributes": attributes,
        }
    }


async def poll_workflow_status(
    client: httpx.AsyncClient,
    execution_id: str,
    expected_status: str,
    timeout_seconds: int = 30,
) -> dict[str, Any]:
    for _ in range(timeout_seconds):
        await asyncio.sleep(1)
        status_response = await client.get(f"/v1/workflows/executions/{execution_id}")
        status_response.raise_for_status()
        status_data = status_response.json()
        status = status_data.get("status")

        if status == expected_status:
            return cast(dict[str, Any], status_data)
        elif status in ["FAILED", "TERMINATED", "TIMED_OUT", "CANCELED", "COMPLETED", "RETRYING_AFTER_ERROR"]:
            if status != expected_status:
                raise RuntimeError(f"Workflow ended with status: {status}, expected: {expected_status}")

    raise TimeoutError(f"Workflow did not reach {expected_status} status within {timeout_seconds}s")


async def poll_pending_inputs(
    client: httpx.AsyncClient,
    execution_id: str,
    expected_count: int = 1,
    timeout_seconds: int = 10,
) -> list[dict[str, Any]]:
    for _ in range(timeout_seconds * 10):
        await asyncio.sleep(0.1)
        try:
            pending_response = await client.post(
                f"/v1/workflows/executions/{execution_id}/queries",
                json={"name": "__get_pending_inputs"},
            )
            pending_response.raise_for_status()
            pending_data = pending_response.json()
            pending_inputs = pending_data.get("result", {}).get("pending_inputs", [])

            if len(pending_inputs) >= expected_count:
                return cast(list[dict[str, Any]], pending_inputs)
        except (httpx.HTTPError, ValueError, KeyError):
            pass

    raise TimeoutError(f"Did not receive {expected_count} pending inputs within {timeout_seconds}s")


async def poll_worker_activity_status(
    client: httpx.AsyncClient,
    expected_active_status: bool,
    workflow_identifier: str | None = None,
    workflow_registration_id: uuid.UUID | None = None,
    timeout_seconds: int = 15,
) -> None:
    if workflow_identifier:
        route = f"/v1/workflows/{workflow_identifier}"
    elif workflow_registration_id:
        route = f"/v1/workflows/registrations/{workflow_registration_id}"
    else:
        raise ValueError("Either workflow_identifier or workflow_registration_id must be provided")

    for _ in range(timeout_seconds):
        response = await client.get(route)
        response.raise_for_status()
        active = response.json().get("workflow", {}).get("active")
        if active == expected_active_status:
            return
        await asyncio.sleep(1)

    raise TimeoutError(f"Worker activity status is still not '{str(expected_active_status)}' after {timeout_seconds}s")


async def submit_workflow_update(
    client: httpx.AsyncClient,
    execution_id: str,
    update_name: str,
    update_input: dict[str, Any],
) -> None:
    update_response = await client.post(
        f"/v1/workflows/executions/{execution_id}/updates",
        json={
            "name": update_name,
            "input": update_input,
        },
    )
    update_response.raise_for_status()


async def execute_workflow(
    client: httpx.AsyncClient,
    workflow_name: str,
    input_data: dict[str, Any],
    task_queue: str | None = DEFAULT_TEST_TASK_QUEUE,
) -> str:
    body: dict[str, Any] = {"input": input_data}
    if task_queue is not None:
        body["task_queue"] = task_queue
    response = await client.post(
        f"/v1/workflows/{workflow_name}/execute",
        json=body,
    )
    response.raise_for_status()
    data = response.json()
    execution_id = data.get("execution_id")
    if not execution_id:
        raise ValueError(f"No execution_id returned: {data}")
    return cast(str, execution_id)


async def wait_for_workflow_events(
    client: httpx.AsyncClient, execution_id: str, max_retries: int = 10
) -> dict[str, Any]:
    for _ in range(max_retries):
        response = await client.get("/v1/workflows/events/list", params={"workflow_exec_id": execution_id, "limit": 1})
        response.raise_for_status()
        events_data = response.json()
        if "events" in events_data and len(events_data["events"]) > 0:
            return cast(dict[str, Any], events_data)
        await asyncio.sleep(0.5)
    raise AssertionError(f"No events found for execution {execution_id} after {max_retries} retries")


async def stream_workflow_events(
    client: httpx.AsyncClient,
    execution_id: str,
    timeout_seconds: float = 30.0,
) -> list[dict[str, Any]]:
    async with asyncio.timeout(timeout_seconds):
        received_events: list[dict[str, Any]] = []
        retry_delay: float = 1  # by default 1 second

        while True:
            start_seq = 0
            if received_events:
                start_seq = received_events[-1]["broker_sequence"] + 1
            async with client.stream(
                "GET",
                "/v1/workflows/events/stream",
                params={"workflow_exec_id": execution_id, "start_seq": start_seq},
                timeout=timeout_seconds,
            ) as stream_response:
                stream_response.raise_for_status()
                async for line in stream_response.aiter_lines():
                    if not line or line.startswith(":"):
                        continue
                    if line.startswith(SSE_DATA_PREFIX):
                        event_data = line[len(SSE_DATA_PREFIX) :].strip()
                        if event_data:
                            event = json.loads(event_data)
                            sse_payload_adapter.validate_python(event)
                            received_events.append(event)
                            event_type = event.get("data", {}).get("event_type")
                            if event_type == WorkflowEventType.WORKFLOW_EXECUTION_COMPLETED:
                                return received_events
                    elif line.startswith(SSE_RETRY_PREFIX):
                        event_retry = line[len(SSE_RETRY_PREFIX) :].strip()
                        if event_retry:
                            try:
                                retry_delay = float(event_retry) / 1000  # the retry is passed in milliseconds
                            except ValueError:
                                continue

            # delay the next retry by the provided amount before retrying
            await asyncio.sleep(retry_delay)


def get_event_types(events: list[dict[str, Any]]) -> list[str]:
    return [e.get("data", {}).get("event_type") for e in events]


def filter_events_by_type(events: list[dict[str, Any]], event_types: list[WorkflowEventType]) -> list[dict[str, Any]]:
    filtered = []
    for e in events:
        event_type = e.get("data", {}).get("event_type")
        if event_type in event_types:
            workflow_event_adapter.validate_python(e.get("data", {}))
            filtered.append(e)
    return filtered


def filter_events_by_custom_task_type(events: list[dict[str, Any]], custom_task_type: str) -> list[dict[str, Any]]:
    filtered = []
    for e in events:
        attrs = e.get("data", {}).get("attributes", {})
        if attrs.get("custom_task_type") == custom_task_type:
            workflow_event_adapter.validate_python(e.get("data", {}))
            filtered.append(e)
    return filtered


def extract_state_from_event(event: dict[str, Any]) -> dict[str, Any] | None:
    validated = sse_payload_adapter.validate_python(event)
    event_data = validated.data
    if isinstance(event_data, (CustomTaskStarted, CustomTaskInProgress, CustomTaskCompleted)):
        payload = event_data.attributes.payload
        if payload.type == "json":
            return cast(dict[str, Any], payload.value)
    return None


def extract_json_patches_from_event(event: dict[str, Any]) -> list[dict[str, Any]]:
    validated = sse_payload_adapter.validate_python(event)
    event_data = validated.data
    if isinstance(event_data, (CustomTaskInProgress, CustomTaskCompleted)):
        payload = event_data.attributes.payload
        if payload.type == "json_patch":
            return [p.model_dump() for p in payload.value]
    return []


async def execute_workflow_and_wait(
    client: httpx.AsyncClient,
    workflow_name: str,
    input_data: dict[str, Any],
    timeout_seconds: int = 30,
    task_queue: str | None = DEFAULT_TEST_TASK_QUEUE,
) -> dict[str, Any]:
    body: dict[str, Any] = {"input": input_data}
    if task_queue is not None:
        body["task_queue"] = task_queue
    response = await client.post(
        f"/v1/workflows/{workflow_name}/execute",
        json=body,
    )
    response.raise_for_status()
    data = response.json()

    execution_id = data.get("execution_id")
    if not execution_id:
        raise ValueError(f"No execution_id returned: {data}")

    for _ in range(timeout_seconds):
        await asyncio.sleep(1)

        status_response = await client.get(f"/v1/workflows/executions/{execution_id}")
        status_response.raise_for_status()
        status_data = status_response.json()

        status = status_data.get("status")
        if status == "COMPLETED":
            return cast(dict[str, Any], status_data)
        elif status in ["FAILED", "TERMINATED", "TIMED_OUT", "CANCELED"]:
            raise RuntimeError(f"Workflow {workflow_name} ended with status: {status}")

    raise TimeoutError(f"Workflow {workflow_name} did not complete within {timeout_seconds}s")


async def poll_trace_events_available(
    client: httpx.AsyncClient, execution_id: str, timeout: float = 30.0, poll_interval: float = 0.5
) -> None:
    start_time = asyncio.get_event_loop().time()

    while True:
        try:
            response = await client.get(f"/v1/workflows/executions/{execution_id}/trace/events")
            if response.status_code == 200:
                return
        except httpx.HTTPStatusError:
            pass

        elapsed = asyncio.get_event_loop().time() - start_time
        if elapsed > timeout:
            raise TimeoutError(f"Trace events not available for execution {execution_id} after {timeout}s")

        await asyncio.sleep(poll_interval)


async def poll_trace_otel_data_available(
    client: httpx.AsyncClient,
    execution_id: str,
    timeout: float = 30.0,
    poll_interval: float = 0.5,
    required_span_prefix: str | None = None,
) -> None:
    """
    Poll until OTEL trace data is available for an execution.

    Args:
        client: HTTP client to use for requests
        execution_id: The workflow execution ID
        timeout: Maximum time to wait in seconds
        poll_interval: Time between polls in seconds
        required_span_prefix: Optional span name prefix to wait for (e.g., "WorkflowReport:")
            If provided, polling continues until a span with this prefix is found.
    """
    start_time = asyncio.get_event_loop().time()

    while True:
        try:
            response = await client.get(f"/v1/workflows/executions/{execution_id}/trace/otel")
            response.raise_for_status()

            trace_data = response.json()
            otel_data = trace_data.get("otel_trace_data", {})

            batches = otel_data.get("batches", [])

            if required_span_prefix is None and len(batches) > 0:
                return

            # Check if required span is present
            if required_span_prefix is not None:
                for batch in batches:
                    for scope_span in batch.get("scopeSpans", []):
                        for span in scope_span.get("spans", []):
                            span_name = span.get("name", "")
                            if span_name.startswith(required_span_prefix):
                                return
        except httpx.HTTPStatusError:
            pass

        elapsed = asyncio.get_event_loop().time() - start_time
        if elapsed > timeout:
            raise TimeoutError(
                f"OTEL trace data not available for execution {execution_id} after {timeout}s"
                + (f" (waiting for span prefix: {required_span_prefix})" if required_span_prefix else "")
            )

        await asyncio.sleep(poll_interval)


async def wait_for_pending_inputs(
    handle: WorkflowHandle,
    expected_count: int = 1,
    timeout: float = 5.0,
    label: str | None = None,
) -> list[dict[str, Any]]:
    start_time = asyncio.get_event_loop().time()

    while True:
        try:
            pending_result = await handle.query("__get_pending_inputs")
            pending_inputs = pending_result["pending_inputs"]

            if label:
                pending_inputs = [inp for inp in pending_inputs if inp.get("label") == label]

            if len(pending_inputs) >= expected_count:
                return cast(list[dict[str, Any]], pending_inputs)

        except Exception:
            pass

        if asyncio.get_event_loop().time() - start_time > timeout:
            raise TimeoutError(
                f"Timeout waiting for {expected_count} pending inputs" + (f" with label '{label}'" if label else "")
            )

        await asyncio.sleep(0.05)


class HeartbeatProcessingParams(BaseModel):
    total_items: int
    items_per_second: int = 10
    get_stuck_at_item: int | None = None
    crash_at_item: int | None = None


class HeartbeatProcessingResult(BaseModel):
    items_processed: int
    heartbeat_count: int
    completed: bool



from pydantic import BaseModel

import mistralai.workflows as workflows


@workflows.activity()
async def activity_with_optional_arg(name: str, id: int = 123) -> str:
    return f"Hello {name}, your ID is {id}"


class ModelArg(BaseModel):
    name: str


@workflows.activity()
async def activity_with_optional_arg_model(name: ModelArg, id: int = 123) -> str:
    return f"Hello {name.name}, your ID is {id}"


@workflows.workflow.define(name="activity-optional-arg-workflow")
class WorkflowCallingActivityWithOptionalArg:
    @workflows.workflow.entrypoint
    async def run(self, name: str) -> str:
        return await activity_with_optional_arg(name)


@workflows.workflow.define(name="activity-optional-arg-workflow-model")
class WorkflowCallingActivityWithOptionalArgModel:
    @workflows.workflow.entrypoint
    async def run(self, name: str) -> str:
        return await activity_with_optional_arg_model(ModelArg(name=name))


@workflows.workflow.define(name="optional-arg-workflow")
class WorkflowWithOptionalArg:
    @workflows.workflow.entrypoint
    async def run(self, name: str, id: int = 123) -> str:
        return f"Hello {name}, your ID is {id}"


@workflows.workflow.define(name="optional-arg-workflow-model")
class WorkflowWithOptionalArgModel:
    @workflows.workflow.entrypoint
    async def run(self, name: ModelArg, id: int = 123) -> str:
        return f"Hello {name.name}, your ID is {id}"



import inspect
from types import NoneType, UnionType
from typing import Any, Callable, Dict, Tuple, Type, TypeVar, Union, get_args, get_origin, get_type_hints

from pydantic import BaseModel, TypeAdapter, ValidationError

from mistralai.workflows.core.definition.validation.schema_generator import (
    generate_pydantic_model_from_params,
    generate_pydantic_model_from_return_type,
)
from mistralai.workflows.core.dependencies.dependency_injector import DependsCls
from mistralai.workflows.core.utils.sandbox import is_instance_or_sandboxed_subclass
from mistralai.workflows.exceptions import ErrorCode, WorkflowsException

T = TypeVar("T", bound=Callable[..., Any])


def extract_origin_type(type_to_extract: Type) -> Type:
    origin = get_origin(type_to_extract)
    args: Tuple[Type, ...] = get_args(type_to_extract)
    if origin is Union and len(args) == 2 and NoneType in args:
        return next(arg for arg in args if arg is not NoneType)
    elif origin is UnionType and len(args) == 2 and NoneType in args:
        return next(arg for arg in args if arg is not NoneType)
    return type_to_extract


def check_is_valid_type(type_to_check: Type, expected_type: Type, allow_optional: bool = False) -> bool:
    if allow_optional:
        type_to_check = extract_origin_type(type_to_check)

    return (inspect.isclass(type_to_check) and issubclass(type_to_check, expected_type)) or (
        type_to_check is expected_type
    )


def get_function_signature_type_hints(func: Callable, is_method: bool) -> Tuple[Dict[str, Type], Type, bool]:
    """Extract type hints and parameter info from a function signature.

    Returns:
        Tuple of (user_params_dict, return_type, has_kwargs)
    """
    type_hints = get_type_hints(func)
    if "return" not in type_hints:
        raise WorkflowsException(
            code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
            message=(
                f"'{func.__name__}' must have a return type annotation, use `-> None` if no return type is expected."
            ),
        )

    sig = inspect.signature(func)

    param_names = list(sig.parameters.keys())

    if is_method:
        if not param_names:
            raise WorkflowsException(
                code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
                message=f"Method '{func.__name__}' has no parameters, expected 'self' or 'cls' as the first.",
            )
        type_hints.pop(param_names[0], None)
        param_names = param_names[1:]

    user_params_dict = {}
    has_kwargs = False
    for param_name in param_names:
        param = sig.parameters[param_name]
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            has_kwargs = True
            continue
        if isinstance(param.default, DependsCls):
            continue

        # Only add to user params if it has a type hint
        if param_name in type_hints:
            user_params_dict[param_name] = type_hints[param_name]
        else:
            raise WorkflowsException(
                code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
                message=f"Parameter '{param_name}' in '{func.__name__}' must have a type annotation.",
            )

    return_type = type_hints["return"]
    return user_params_dict, return_type, has_kwargs


def raise_if_function_has_invalid_signature(
    func: Callable, is_method: bool = False, allow_kwargs: bool = False
) -> None:
    """Validate function signature. Raises if invalid."""
    if not inspect.iscoroutinefunction(func):
        raise WorkflowsException(
            code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
            message=f"'{func.__name__}' must be async function. Use `async def`.",
        )

    user_params_dict, return_type, has_kwargs = get_function_signature_type_hints(func, is_method=is_method)

    if has_kwargs and not allow_kwargs:
        raise WorkflowsException(
            code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
            message=f"'{func.__name__}' has **kwargs which is not supported for this handler type.",
        )

    try:
        generate_pydantic_model_from_params(func.__name__, user_params_dict, func=func)
    except Exception as e:
        raise WorkflowsException(
            code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
            message=f"Cannot generate Pydantic model from parameters of '{func.__name__}': {e}",
        ) from e

    try:
        generate_pydantic_model_from_return_type(func.__name__, return_type)
    except Exception as e:
        raise WorkflowsException(
            code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
            message=f"Cannot generate Pydantic model from return type of '{func.__name__}': {e}",
        ) from e


def raise_if_function_has_invalid_usage(
    func: Callable, args: Tuple[Any, ...], kwargs: Dict[str, Any], is_method: bool = False
) -> None:
    user_params_dict, _, has_kwargs = get_function_signature_type_hints(func, is_method=is_method)

    if is_method:
        args = args[1:]

    sig = inspect.signature(func)
    param_names = list(sig.parameters.keys())
    if is_method:
        param_names = param_names[1:]

    # Build list of required param names (in order) and total param count
    required_params: list[str] = []
    explicit_params: list[str] = []  # Non-VAR_KEYWORD, non-Depends params
    for param_name in param_names:
        param = sig.parameters[param_name]
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            continue
        if isinstance(param.default, DependsCls):
            continue
        explicit_params.append(param_name)
        if param.default is inspect.Parameter.empty:
            required_params.append(param_name)

    total_param_count = len(explicit_params)
    provided_positional = len(args)

    # Check for unknown kwargs when function doesn't have **kwargs
    if not has_kwargs:
        unknown_kwargs = set(kwargs.keys()) - set(explicit_params)
        if unknown_kwargs:
            raise WorkflowsException(
                code=ErrorCode.INVALID_ARGUMENTS_ERROR,
                message=(f"'{func.__name__}' got unexpected keyword arguments: {sorted(unknown_kwargs)}."),
            )

    # Check for duplicate args (provided both positionally and as kwarg)
    for i, param_name in enumerate(explicit_params):
        if i < provided_positional and param_name in kwargs:
            raise WorkflowsException(
                code=ErrorCode.INVALID_ARGUMENTS_ERROR,
                message=(
                    f"'{func.__name__}' got multiple values for parameter '{param_name}': "
                    f"'{args[i]}' and '{kwargs[param_name]}'"
                ),
            )

    # Check too many positional args
    if provided_positional > total_param_count:
        raise WorkflowsException(
            code=ErrorCode.INVALID_ARGUMENTS_ERROR,
            message=(
                f"'{func.__name__}' expects at most {total_param_count} positional parameters. "
                f"Found: {provided_positional} positional arguments."
            ),
        )

    # Check each required param is provided (by position or by name in kwargs)
    for i, param_name in enumerate(required_params):
        param_index = explicit_params.index(param_name)
        provided_by_position = param_index < provided_positional
        provided_by_kwarg = param_name in kwargs
        if not provided_by_position and not provided_by_kwarg:
            raise WorkflowsException(
                code=ErrorCode.INVALID_ARGUMENTS_ERROR,
                message=f"'{func.__name__}' missing required parameter: '{param_name}'.",
            )

    for i, (param_name, param_type) in enumerate(user_params_dict.items()):
        if i < len(args):
            arg = args[i]
        elif param_name in kwargs:
            arg = kwargs[param_name]
        else:
            continue

        if inspect.isclass(param_type) and issubclass(param_type, BaseModel):
            if not is_instance_or_sandboxed_subclass(arg, param_type):
                raise WorkflowsException(
                    code=ErrorCode.INVALID_ARGUMENTS_ERROR,
                    message=(
                        f"Parameter '{param_name}' in '{func.__name__}' should be of type '{param_type}'. "
                        f"Found: '{type(arg)}'."
                    ),
                )


def raise_if_function_has_invalid_return_type(func: Callable, return_value: Any, is_method: bool = False) -> None:
    _, return_type, _ = get_function_signature_type_hints(func, is_method=is_method)

    try:
        adapter = TypeAdapter(return_type)
        adapter.validate_python(return_value, strict=True)
    except ValidationError as e:
        raise WorkflowsException(
            code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
            message=f"'{func.__name__}' return type validation failed: {e}",
        ) from e


def _validate_handler_signature(
    func: Callable,
    is_method: bool,
    error_code: ErrorCode,
    handler_type: str,
    validate_return_type: bool = False,
    require_non_none_return: bool = False,
) -> None:
    user_params_dict, return_type, has_kwargs = get_function_signature_type_hints(func, is_method=is_method)

    try:
        generate_pydantic_model_from_params(func.__name__, user_params_dict, func=func, allow_extra=has_kwargs)
    except Exception as e:
        raise WorkflowsException(
            code=error_code,
            message=f"{handler_type} '{func.__name__}' has invalid parameters for schema generation: {e}",
        ) from e

    if require_non_none_return and (return_type is NoneType or return_type is type(None)):
        raise WorkflowsException(
            code=error_code,
            message=f"{handler_type} '{func.__name__}' must have a return type annotation other than None.",
        )

    if validate_return_type:
        try:
            generate_pydantic_model_from_return_type(func.__name__, return_type)
        except Exception as e:
            raise WorkflowsException(
                code=error_code,
                message=f"{handler_type} '{func.__name__}' has invalid return type for schema generation: {e}",
            ) from e


def validate_signal_handler_signature(func: Callable, is_method: bool = True) -> None:
    _validate_handler_signature(
        func,
        is_method=is_method,
        error_code=ErrorCode.WORKFLOW_SIGNAL_DEFINITION_ERROR,
        handler_type="Signal",
    )


def validate_query_handler_signature(func: Callable, is_method: bool = True) -> None:
    if inspect.iscoroutinefunction(func):
        raise WorkflowsException(
            code=ErrorCode.REJECTED_QUERY_ERROR,
            message=f"Query '{func.__name__}' must be a synchronous function (def), not async def.",
        )

    _validate_handler_signature(
        func,
        is_method=is_method,
        error_code=ErrorCode.REJECTED_QUERY_ERROR,
        handler_type="Query",
        validate_return_type=True,
        require_non_none_return=True,
    )


def validate_update_handler_signature(func: Callable, is_method: bool = True) -> None:
    _validate_handler_signature(
        func,
        is_method=is_method,
        error_code=ErrorCode.WORKFLOW_UPDATE_DEFINITION_ERROR,
        handler_type="Update",
        validate_return_type=True,
        require_non_none_return=True,
    )



from pydantic import BaseModel

import mistralai.workflows as workflows
from mistralai.workflows.core.temporal.context_handler_interceptor import retrieve_context


class EchoResult(BaseModel):
    execution_token: str | None


class EmptyParams(BaseModel):
    pass


class ModelInput(BaseModel):
    message: str
    count: int = 0


@workflows.workflow.define(
    name="example-execution-token-echo-workflow",
    workflow_description="Echo workflow that returns its execution token",
)
class ExecutionTokenEchoWorkflow:
    @workflows.workflow.entrypoint
    async def run(self) -> EchoResult:
        context = retrieve_context()
        token = context.execution_token if context else None
        return EchoResult(execution_token=token)


@workflows.workflow.define(
    name="example-execution-token-echo-parent-workflow",
    workflow_description="Parent workflow that delegates to the echo child and returns its result",
)
class ExecutionTokenEchoParentWorkflow:
    @workflows.workflow.entrypoint
    async def run(self) -> EchoResult:
        result: EchoResult = await workflows.workflow.execute_workflow(ExecutionTokenEchoWorkflow, params=EmptyParams())
        return result


class DictEchoParams(BaseModel):
    message: str = ""


@workflows.workflow.define(name="example-execution-token-dict-echo-workflow")
class ExecutionTokenDictEchoWorkflow:
    @workflows.workflow.entrypoint
    async def run(self, message: str = "") -> EchoResult:
        context = retrieve_context()
        token = context.execution_token if context else None
        return EchoResult(execution_token=token)


@workflows.workflow.define(name="example-execution-token-dict-echo-parent-workflow")
class ExecutionTokenDictEchoParentWorkflow:
    @workflows.workflow.entrypoint
    async def run(self, message: str = "") -> EchoResult:
        result: EchoResult = await workflows.workflow.execute_workflow(
            ExecutionTokenDictEchoWorkflow, params=DictEchoParams(message=message)
        )
        return result


@workflows.workflow.define(name="example-execution-token-model-echo-workflow")
class ExecutionTokenModelEchoWorkflow:
    @workflows.workflow.entrypoint
    async def run(self, params: ModelInput) -> EchoResult:
        context = retrieve_context()
        token = context.execution_token if context else None
        return EchoResult(execution_token=token)


@workflows.activity()
async def get_execution_token_activity() -> EchoResult:
    context = retrieve_context()
    token = context.execution_token if context else None
    return EchoResult(execution_token=token)


@workflows.workflow.define(
    name="example-execution-token-activity-echo-workflow",
    workflow_description="Workflow that retrieves its execution token from within an activity",
)
class ExecutionTokenActivityEchoWorkflow:
    @workflows.workflow.entrypoint
    async def run(self) -> EchoResult:
        return await get_execution_token_activity()



from enum import Enum
from typing import Any, Optional, Union

import pytest
from pydantic import BaseModel

from mistralai.workflows.core.temporal.context_handler_interceptor import _validate_activity_args


class SimpleModel(BaseModel):
    name: str
    value: int


class NestedInner(BaseModel):
    x: int
    y: int


class NestedModel(BaseModel):
    label: str
    inner: NestedInner


class ModelWithDefaults(BaseModel):
    name: str
    score: float = 0.0


class Color(str, Enum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"


class ModelWithOptionalField(BaseModel):
    name: str
    tag: Optional[str] = None


def fn_no_params() -> None:
    pass


def fn_no_hints(a, b):  # noqa: ANN001, ANN201
    pass


def fn_all_any(a: Any, b: Any) -> None:
    pass


def fn_simple(name: str, value: int) -> None:
    pass


def fn_single_model(model: SimpleModel) -> None:
    pass


def fn_nested_model(data: NestedModel) -> None:
    pass


def fn_model_with_defaults(data: ModelWithDefaults) -> None:
    pass


def fn_optional_str(name: str, tag: Optional[str] = None) -> None:
    pass


def fn_union(value: Union[int, str]) -> None:
    pass


def fn_pipe_union(value: int | str) -> None:
    pass


def fn_list_of_int(items: list[int]) -> None:
    pass


def fn_dict_str_int(mapping: dict[str, int]) -> None:
    pass


def fn_enum_param(color: Color) -> None:
    pass


def fn_with_default(name: str, count: int = 10) -> None:
    pass


def fn_mixed(name: str, model: SimpleModel, flag: bool) -> None:
    pass


def fn_multiple_models(a: SimpleModel, b: NestedModel) -> None:
    pass


def fn_optional_model(name: str, model: Optional[SimpleModel] = None) -> None:
    pass


def fn_list_of_models(items: list[SimpleModel]) -> None:
    pass


def fn_bool(flag: bool) -> None:
    pass


def fn_model_with_opt_field(m: ModelWithOptionalField) -> None:
    pass


def fn_complex(
    name: str,
    model: SimpleModel,
    tags: list[str],
    metadata: dict[str, Any],
    count: int = 5,
) -> None:
    pass


class _Svc:
    def process(self, data: SimpleModel) -> None:
        pass


_svc_process = _Svc().process


def fn_bad_forward_ref(x: "NonExistentType") -> None:  # noqa: F821
    pass


@pytest.mark.parametrize(
    "fn, args, expected",
    [
        pytest.param(fn_no_params, [], [], id="no_params_no_args"),
        pytest.param(fn_simple, ["hello", 42], ["hello", 42], id="simple_matching_types"),
        pytest.param(
            fn_single_model,
            [{"name": "test", "value": 42}],
            [SimpleModel(name="test", value=42)],
            id="dict_to_simple_model",
        ),
        pytest.param(
            fn_nested_model,
            [{"label": "point", "inner": {"x": 1, "y": 2}}],
            [NestedModel(label="point", inner=NestedInner(x=1, y=2))],
            id="dict_to_nested_model",
        ),
        pytest.param(
            fn_model_with_defaults,
            [{"name": "hi"}],
            [ModelWithDefaults(name="hi")],
            id="model_defaults_partial",
        ),
        pytest.param(
            fn_model_with_defaults,
            [{"name": "hi", "score": 9.5}],
            [ModelWithDefaults(name="hi", score=9.5)],
            id="model_defaults_full",
        ),
        pytest.param(
            fn_multiple_models,
            [{"name": "a", "value": 1}, {"label": "b", "inner": {"x": 0, "y": 0}}],
            [SimpleModel(name="a", value=1), NestedModel(label="b", inner=NestedInner(x=0, y=0))],
            id="multiple_models",
        ),
        pytest.param(
            fn_list_of_models,
            [[{"name": "a", "value": 1}, {"name": "b", "value": 2}]],
            [[SimpleModel(name="a", value=1), SimpleModel(name="b", value=2)]],
            id="list_of_models",
        ),
        pytest.param(
            fn_optional_model,
            ["hello", None],
            ["hello", None],
            id="optional_model_none",
        ),
        pytest.param(
            fn_optional_model,
            ["hello", {"name": "x", "value": 1}],
            ["hello", SimpleModel(name="x", value=1)],
            id="optional_model_dict",
        ),
        pytest.param(fn_no_hints, ["a", 2], ["a", 2], id="no_type_hints"),
        pytest.param(fn_all_any, ["hello", {"x": 1}], ["hello", {"x": 1}], id="any_type_hints"),
        pytest.param(fn_optional_str, ["hello", "world"], ["hello", "world"], id="optional_str_value"),
        pytest.param(fn_optional_str, ["hello", None], ["hello", None], id="optional_str_none"),
        pytest.param(fn_union, [42], [42], id="union_int"),
        pytest.param(fn_union, ["hello"], ["hello"], id="union_str"),
        pytest.param(fn_pipe_union, [42], [42], id="pipe_union"),
        pytest.param(fn_enum_param, ["red"], [Color.RED], id="enum_from_str"),
        pytest.param(fn_enum_param, [Color.GREEN], [Color.GREEN], id="enum_already_enum"),
        pytest.param(fn_list_of_int, [[1, 2, 3]], [[1, 2, 3]], id="list_of_int"),
        pytest.param(fn_dict_str_int, [{"a": 1, "b": 2}], [{"a": 1, "b": 2}], id="dict_str_int"),
        pytest.param(fn_with_default, ["hello"], ["hello"], id="fewer_args_than_params"),
        pytest.param(fn_simple, ["a", 1, "extra"], ["a", 1, "extra"], id="more_args_than_params"),
        pytest.param(
            fn_single_model,
            [{"name": "x", "value": 1}, "extra_arg"],
            [SimpleModel(name="x", value=1), "extra_arg"],
            id="more_args_with_model",
        ),
        pytest.param(fn_with_default, [], [], id="empty_args_with_defaults"),
        pytest.param(
            fn_complex,
            ["test_name", {"name": "m", "value": 10}, ["tag1", "tag2"], {"key": "val"}, 3],
            ["test_name", SimpleModel(name="m", value=10), ["tag1", "tag2"], {"key": "val"}, 3],
            id="complex_all_args",
        ),
        pytest.param(
            fn_complex,
            ["test_name", {"name": "m", "value": 10}, ["tag1"], {"k": 1}],
            ["test_name", SimpleModel(name="m", value=10), ["tag1"], {"k": 1}],
            id="complex_optional_omitted",
        ),
        pytest.param(
            fn_model_with_opt_field,
            [{"name": "test"}],
            [ModelWithOptionalField(name="test")],
            id="optional_field_missing",
        ),
        pytest.param(
            fn_model_with_opt_field,
            [{"name": "test", "tag": "v1"}],
            [ModelWithOptionalField(name="test", tag="v1")],
            id="optional_field_provided",
        ),
        pytest.param(
            fn_single_model,
            [SimpleModel(name="ok", value=1)],
            [SimpleModel(name="ok", value=1)],
            id="already_instantiated_model",
        ),
        pytest.param(fn_simple, ["hello", 3.0], ["hello", 3], id="int_from_float"),
        pytest.param(fn_bool, [True], [True], id="bool_coercion"),
        pytest.param(
            fn_single_model,
            [{"wrong_field": "oops"}],
            [{"wrong_field": "oops"}],
            id="invalid_dict_for_model_keeps_original",
        ),
        pytest.param(
            fn_list_of_int,
            ["not_a_list"],
            ["not_a_list"],
            id="wrong_type_keeps_original",
        ),
        pytest.param(
            fn_mixed,
            ["hello", {"bad": "dict"}, "not_bool"],
            ["hello", {"bad": "dict"}, "not_bool"],
            id="partial_validation_failure",
        ),
        pytest.param(
            fn_simple,
            [None, None],
            [None, None],
            id="none_for_non_optional_keeps_original",
        ),
        pytest.param(lambda a, b: None, [1, 2], [1, 2], id="lambda_no_hints"),
        pytest.param(
            _svc_process,
            [{"name": "a", "value": 1}],
            [SimpleModel(name="a", value=1)],
            id="bound_method",
        ),
        pytest.param(fn_bad_forward_ref, ["hello"], ["hello"], id="unresolvable_forward_ref"),
    ],
)
def test_validates_args(fn: Any, args: list, expected: list) -> None:
    result = _validate_activity_args(fn, args)
    assert list(result) == expected



from mistralai.workflows.models import EncodedPayload, EncodedPayloadOptions, NetworkEncodedInput, WorkflowContext


class TestNetworkEncodedInput:
    def test_from_data(self):
        encoded = NetworkEncodedInput.from_data(b"test", [EncodedPayloadOptions.ENCRYPTED])
        assert encoded.get_payload() == b"test"
        assert encoded.encoding_options == [EncodedPayloadOptions.ENCRYPTED]

    def test_from_encoded_payload(self):
        payload = EncodedPayload(
            payload=b"test",
            encoding_options=[EncodedPayloadOptions.OFFLOADED],
            context=WorkflowContext(namespace="ns", execution_id="exec"),
        )
        encoded = NetworkEncodedInput.from_encoded_payload(payload)
        assert encoded.get_payload() == b"test"
        assert encoded.encoding_options == [EncodedPayloadOptions.OFFLOADED]

    def test_to_encoded_payload(self):
        encoded = NetworkEncodedInput.from_data(b"test", [EncodedPayloadOptions.ENCRYPTED])
        payload = encoded.to_encoded_payload(namespace="ns", execution_id="exec")
        assert payload.payload == b"test"
        assert payload.encoding_options == [EncodedPayloadOptions.ENCRYPTED]
        assert payload.context.namespace == "ns"
        assert payload.context.execution_id == "exec"

    def test_get_payload(self):
        encoded = NetworkEncodedInput.from_data(b"test", [])
        assert encoded.get_payload() == b"test"

    def test_roundtrip(self):
        original = EncodedPayload(
            payload=b'{"key": "value"}',
            encoding_options=[EncodedPayloadOptions.ENCRYPTED, EncodedPayloadOptions.OFFLOADED],
            context=WorkflowContext(namespace="ns", execution_id="exec"),
        )
        encoded = NetworkEncodedInput.from_encoded_payload(original)
        restored = encoded.to_encoded_payload(namespace="ns", execution_id="exec")
        assert restored.payload == original.payload
        assert restored.encoding_options == original.encoding_options



from typing import Any
from unittest.mock import AsyncMock

import pytest
from pydantic import BaseModel
from temporalio.testing import WorkflowEnvironment
from temporalio.worker import Replayer
from temporalio.worker.workflow_sandbox import SandboxedWorkflowRunner

from mistralai.workflows.client import WorkflowsClient
from mistralai.workflows.core.activity import activity
from mistralai.workflows.core.events.event_context import EventContext
from mistralai.workflows.core.events.event_interceptor import EventInterceptor
from mistralai.workflows.core.sandbox import SANDBOX_RESTRICTIONS
from mistralai.workflows.core.task import task
from mistralai.workflows.core.workflow import workflow
from mistralai.workflows.protocol.v1.events import WorkflowEventType

from .utils import create_test_worker_with_events


class TaskState(BaseModel):
    progress: int = 0


@activity(name="replay_test_step")
async def do_step(name: str) -> str:
    return f"{name}_done"


@workflow.define(name="replay_test_task_workflow")
class TaskWithActivitiesWorkflow:
    @workflow.entrypoint
    async def run(self) -> dict[str, Any]:
        async with task("work", state=TaskState()) as t:
            await t.update_state({"progress": 50})
            result = await do_step("step1")
            await t.update_state({"progress": 100})
            return {"result": result, "progress": t.state.progress}


class TestTaskReplayDeterminism:
    @pytest.mark.asyncio
    async def test_events_published_during_execution_not_during_replay(self, temporal_env: WorkflowEnvironment) -> None:
        """Events published during first run, none during replay, no nondeterminism."""
        execution_events: list[Any] = []
        replay_events: list[Any] = []

        mock_client = AsyncMock(spec=WorkflowsClient)
        mock_client.send_event = AsyncMock(side_effect=lambda e: execution_events.append(e))

        async with EventContext(mock_client):
            async with create_test_worker_with_events(
                temporal_env,
                workflows=[TaskWithActivitiesWorkflow],
                activities=[do_step],
            ):
                handle = await temporal_env.client.start_workflow(
                    "replay_test_task_workflow",
                    id="replay-test",
                    task_queue="test-task-queue",
                )
                await handle.result()
                history = await handle.fetch_history()

        assert len(execution_events) > 0, "Events should be published during execution"

        event_types = {e.event_type for e in execution_events}
        assert WorkflowEventType.WORKFLOW_EXECUTION_STARTED in event_types
        assert WorkflowEventType.WORKFLOW_EXECUTION_COMPLETED in event_types
        assert WorkflowEventType.CUSTOM_TASK_STARTED in event_types
        assert WorkflowEventType.CUSTOM_TASK_COMPLETED in event_types

        mock_client_replay = AsyncMock(spec=WorkflowsClient)
        mock_client_replay.send_event = AsyncMock(side_effect=lambda e: replay_events.append(e))

        async with EventContext(mock_client_replay):
            replayer = Replayer(
                workflows=[TaskWithActivitiesWorkflow],
                interceptors=[EventInterceptor()],
                workflow_runner=SandboxedWorkflowRunner(
                    restrictions=SANDBOX_RESTRICTIONS,
                ),
            )
            await replayer.replay_workflow(history)

        assert len(replay_events) == 0, "No events should be published during replay"
